from test.testbase import DbIsolatedTest, run_as_sync
from pymysql.err import IntegrityError


from db.python.connect import NotFoundError
from db.python.layers.sample import SampleLayer
from db.python.layers.sequence import SampleSequenceLayer, SequenceType, SequenceStatus
from models.models.sequence import SampleSequencing
from models.enums import SampleType


class TestSequence(DbIsolatedTest):
    """Test sequence class"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.slayer = SampleLayer(self.connection)
        self.seqlayer = SampleSequenceLayer(self.connection)
        self.external_sample_id = 'TESTING001'

        # Create new sample
        self.sample_id_raw = await self.slayer.insert_sample(
            self.external_sample_id,
            SampleType.BLOOD,
            active=True,
            meta={'Testing': 'test_sequence'},
        )

    @run_as_sync
    async def test_not_found_sequence(self):
        """
        Test the NotFoundError when getting an invalid sequence ID
        """

        @run_as_sync
        async def get():
            return await self.seqlayer.get_sequence_by_id(999, check_project_id=False)

        self.assertRaises(NotFoundError, get)

    @run_as_sync
    async def test_insert_sequence(self):
        """
        Test inserting a sequence, and check all values are inserted correctly
        """
        external_ids = {'default': 'SEQ01', 'collaborator2': 'CBSEQ_1'}
        meta = {'1': 1, 'nested': {'nested': 'dict'}, 'alpha': ['b', 'e', 't']}
        seq_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta=meta,
            external_ids=external_ids,
        )

        sequence = await self.seqlayer.get_sequence_by_id(
            sequence_id=seq_id, check_project_id=False
        )

        self.assertEqual(seq_id, sequence.id)
        self.assertEqual(self.sample_id_raw, int(sequence.sample_id))
        self.assertEqual('genome', sequence.type.value)
        self.assertEqual('uploaded', sequence.status.value)
        self.assertDictEqual(external_ids, sequence.external_ids)
        self.assertDictEqual(meta, sequence.meta)

    @run_as_sync
    async def test_insert_sequences_for_each_type(self):
        """
        Test inserting a sequence, and check all values are inserted correctly
        """
        meta = {'1': 1, 'nested': {'nested': 'dict'}, 'alpha': ['b', 'e', 't']}
        seq_ids = await self.seqlayer.insert_many_sequencing(
            [
                SampleSequencing(
                    external_ids={},
                    sample_id=self.sample_id_raw,
                    type=stype,
                    meta=meta,
                    status=SequenceStatus.COMPLETED_SEQUENCING,
                )
                for stype in SequenceType
            ]
        )

        inserted_types_rows = await self.connection.connection.fetch_all(
            'SELECT type FROM sample_sequencing WHERE id in :ids', {'ids': seq_ids}
        )
        inserted_types = set(r['type'] for r in inserted_types_rows)

        self.assertEqual(len(SequenceType), len(seq_ids))
        self.assertSetEqual(set(t.value for t in SequenceType), inserted_types)

    @run_as_sync
    async def test_clashing_external_ids(self):
        """Test that should fail when 2nd sequence is inserted with same external_id"""
        external_ids = {'default': 'clashing'}
        await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids=external_ids,
        )

        @run_as_sync
        async def _insert_failing_sequence():
            return await self.seqlayer.insert_sequencing(
                sample_id=self.sample_id_raw,
                sequence_type=SequenceType.GENOME,
                status=SequenceStatus.UPLOADED,
                sequence_meta={},
                external_ids=external_ids,
            )

        _n_sequences_query = 'SELECT COUNT(*) from sample_sequencing'
        self.assertEqual(
            1, await self.connection.connection.fetch_val(_n_sequences_query)
        )
        self.assertRaises(IntegrityError, _insert_failing_sequence)
        # make sure the transaction unwinds the insert second sequence if the external_id clashes
        self.assertEqual(
            1, await self.connection.connection.fetch_val(_n_sequences_query)
        )

    @run_as_sync
    async def test_insert_clashing_external_ids_multiple(self):
        """
        Test inserting a sequence, and check all values are inserted correctly
        """
        external_ids = {'default': 'clashing'}

        @run_as_sync
        async def _insert_clashing():
            return await self.seqlayer.insert_many_sequencing(
                [
                    SampleSequencing(
                        # both get the same external_ids
                        external_ids=external_ids,
                        sample_id=self.sample_id_raw,
                        type=SequenceType.EXOME,
                        meta={},
                        status=SequenceStatus.COMPLETED_SEQUENCING,
                    )
                    for _ in range(2)
                ]
            )

        _n_sequences_query = 'SELECT COUNT(*) from sample_sequencing'

        self.assertEqual(
            0, await self.connection.connection.fetch_val(_n_sequences_query)
        )
        self.assertRaises(IntegrityError, _insert_clashing)
        self.assertEqual(
            0, await self.connection.connection.fetch_val(_n_sequences_query)
        )

    @run_as_sync
    async def test_getting_sequence_by_external_id(self):
        """
        Test get differences sequences by multiple IDs
        """
        seq1_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids={'default': 'SEQ01', 'other': 'EXT_SEQ1'},
        )
        seq2_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.EXOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids={'default': 'SEQ02'},
        )

        self.assertEqual(
            seq1_id, (await self.seqlayer.get_sequence_by_external_id('SEQ01')).id
        )
        self.assertEqual(
            seq1_id, (await self.seqlayer.get_sequence_by_external_id('EXT_SEQ1')).id
        )
        seq2 = await self.seqlayer.get_sequence_by_external_id('SEQ02')
        self.assertEqual(seq2_id, seq2.id)

    @run_as_sync
    async def test_get_sequences_by_sample_id(self):
        """Get many sequences by sample ID"""
        seq1_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids={},
        )
        seq2_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.EXOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids={},
        )

        by_type = await self.seqlayer.get_sequence_ids_for_sample_ids_by_type(
            [self.sample_id_raw], check_project_ids=False
        )

        self.assertIn(self.sample_id_raw, by_type)
        self.assertEqual(1, len(by_type))

        self.assertListEqual(
            [seq1_id], by_type[self.sample_id_raw][SequenceType.GENOME]
        )
        self.assertListEqual([seq2_id], by_type[self.sample_id_raw][SequenceType.EXOME])

    @run_as_sync
    async def test_query(self):
        """Test query_sequences in different combinations"""
        sample_id_for_test = await self.slayer.insert_sample(
            'SAM_TEST_QUERY',
            SampleType.BLOOD,
            active=True,
            meta={'collection-year': '2022'},
        )

        seq1_id = await self.seqlayer.insert_sequencing(
            sample_id=sample_id_for_test,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={'unique': 'a', 'common': 'common'},
            external_ids={'default': 'SEQ01'},
        )
        seq2_id = await self.seqlayer.insert_sequencing(
            sample_id=sample_id_for_test,
            sequence_type=SequenceType.EXOME,
            status=SequenceStatus.RECEIVED,
            sequence_meta={'unique': 'b', 'common': 'common'},
            external_ids={'default': 'SEQ02'},
        )

        async def search_result_to_ids(**query):
            seqs = await self.seqlayer.get_sequences_by(
                **query, project_ids=[self.project_id]
            )
            return {s.id for s in seqs}

        # sample_ids
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(sample_ids=[sample_id_for_test]),
        )
        self.assertSetEqual(set(), await search_result_to_ids(sample_ids=[9_999_999]))

        # external sequence IDs
        self.assertSetEqual(
            {seq1_id}, await search_result_to_ids(external_sequence_ids=['SEQ01'])
        )
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(external_sequence_ids=['SEQ01', 'SEQ02']),
        )

        # seq_meta
        self.assertSetEqual(
            {seq2_id}, await search_result_to_ids(seq_meta={'unique': 'b'})
        )
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(seq_meta={'common': 'common'}),
        )

        # sample meta
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(sample_meta={'collection-year': '2022'}),
        )
        self.assertSetEqual(
            set(), await search_result_to_ids(sample_meta={'unknown_key': '2022'})
        )

        # sequence types
        self.assertSetEqual(
            {seq1_id}, await search_result_to_ids(types=[SequenceType.GENOME])
        )
        self.assertSetEqual(
            {seq2_id}, await search_result_to_ids(types=[SequenceType.EXOME])
        )
        self.assertSetEqual(
            {seq1_id, seq2_id},
            await search_result_to_ids(types=[SequenceType.GENOME, SequenceType.EXOME]),
        )

        # combination
        self.assertSetEqual(
            {seq2_id},
            await search_result_to_ids(
                sample_meta={'collection-year': '2022'}, external_sequence_ids=['SEQ02']
            ),
        )
        self.assertSetEqual(
            {seq1_id},
            await search_result_to_ids(
                external_sequence_ids=['SEQ01'], types=[SequenceType.GENOME]
            ),
        )
        self.assertSetEqual(
            set(),
            await search_result_to_ids(
                external_sequence_ids=['SEQ01'], types=[SequenceType.EXOME]
            ),
        )

    @run_as_sync
    async def test_update(self):
        """Test updating a sequence, and all fields are updated correctly"""
        seq_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.RECEIVED,
            sequence_meta={'a': 1, 'b': 2},
            external_ids={
                'default': 'SEQ01',
                'untouched': 'UTC+1',
                'to_delete': 'VALUE',
            },
        )

        await self.seqlayer.update_sequence(
            seq_id,
            external_ids={'default': 'NSQ_01', 'ext': 'EXTSEQ01', 'to_delete': None},
            status=SequenceStatus.UPLOADED,
            meta={'a': 2, 'c': True},
        )

        update_sequence = await self.seqlayer.get_sequence_by_id(
            sequence_id=seq_id, check_project_id=False
        )

        self.assertEqual(seq_id, update_sequence.id)
        self.assertEqual(self.sample_id_raw, int(update_sequence.sample_id))
        self.assertEqual('genome', update_sequence.type.value)
        self.assertEqual('uploaded', update_sequence.status.value)
        self.assertDictEqual(
            {'default': 'NSQ_01', 'ext': 'EXTSEQ01', 'untouched': 'UTC+1'},
            update_sequence.external_ids,
        )
        self.assertDictEqual({'a': 2, 'b': 2, 'c': True}, update_sequence.meta)

    @run_as_sync
    async def test_update_status(self):
        """
        Test update all sequence statuses
        """
        seq_id = await self.seqlayer.insert_sequencing(
            sample_id=self.sample_id_raw,
            sequence_type=SequenceType.GENOME,
            status=SequenceStatus.UPLOADED,
            sequence_meta={},
            external_ids={},
        )

        # cycle through all statuses, and check that works
        for status in SequenceStatus:
            await self.seqlayer.update_status(seq_id, status, check_project_id=False)
            status_to_check = await self.connection.connection.fetch_one(
                'SELECT status FROM sample_sequencing WHERE id = :id', {'id': seq_id}
            )
            self.assertEqual(status.value, status_to_check['status'])



import unittest
from io import StringIO
from unittest.mock import patch
from test.testbase import run_as_sync

from sample_metadata.parser.generic_metadata_parser import GenericMetadataParser


class TestParseGenericMetadata(unittest.TestCase):
    """Test the GenericMetadataParser"""

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('os.path.getsize')
    async def test_key_map(
        self, mock_stat_size, mock_get_sequence_ids, mock_get_sample_id
    ):
        """
        Test the flexible key map + other options
        """
        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}
        mock_stat_size.return_value = 111

        rows = [
            'sample id,filenames',
            '<sample-id>,<sample-id>-R1.fastq.gz',
            '<sample-id>,<sample-id>-R2.fastq.gz',
        ]

        parser = GenericMetadataParser(
            search_locations=['.'],
            key_map={
                'fn': ['filenames', 'filename'],
                'sample': ['sample id', 'sample_id'],
            },
            ignore_extra_keys=False,
            reads_column='fn',
            sample_name_column='sample',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            '<sample-id>-R1.fastq.gz': 'gs://<sample-id>-R1.fastq.gz',
            '<sample-id>-R2.fastq.gz': 'gs://<sample-id>-R2.fastq.gz',
        }

        resp = await parser.parse_manifest(
            StringIO('\n'.join(rows)), delimiter=',', dry_run=True
        )

        self.assertEqual(1, len(resp['samples']['insert']))
        self.assertEqual(1, len(resp['sequences']['insert']))
        self.assertEqual(0, len(resp['samples']['update']))
        self.assertEqual(0, len(resp['sequences']['update']))

        parser.ignore_extra_keys = False
        rows = [
            'sample id,filenames,extra',
            '<sample-id>,<sample-id>-R1.fastq.gz,extra',
            '<sample-id>,<sample-id>-R2.fastq.gz,read-all-about-it',
        ]

        try:
            _ = await parser.parse_manifest(
                StringIO('\n'.join(rows)), delimiter=',', dry_run=True
            )
        except ValueError as e:
            self.assertEqual(
                'Key "extra" not found in provided key map: fn, sample', str(e)
            )

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.parser.cloudhelper.AnyPath')
    async def test_single_row(
        self, mock_any_path, mock_get_sequence_ids, mock_get_sample_id
    ):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_sequence_ids_for_sample_ids_by_type
        """
        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}

        # new magic mocks across AnyPath
        mock_any_path.return_value.stat.return_value.st_size = 111
        mock_any_path.return_value.exists.return_value = False

        rows = [
            'GVCF\tCRAM\tSampleId\tsample.flowcell_lane\tsample.platform\tsample.centre\tsample.reference_genome\traw_data.FREEMIX\traw_data.PCT_CHIMERAS\traw_data.MEDIAN_INSERT_SIZE\traw_data.MEDIAN_COVERAGE',
            '<sample-id>.g.vcf.gz\t<sample-id>.bam\t<sample-id>\tHK7NFCCXX.1\tILLUMINA\tKCCG\thg38\t0.01\t0.01\t400\t30',
        ]
        parser = GenericMetadataParser(
            search_locations=[],
            sample_name_column='SampleId',
            participant_meta_map={},
            sample_meta_map={'sample.centre': 'centre'},
            sequence_meta_map={
                'raw_data.FREEMIX': 'qc.freemix',
                'raw_data.PCT_CHIMERAS': 'qc.pct_chimeras',
                'raw_data.MEDIAN_INSERT_SIZE': 'qc.median_insert_size',
                'raw_data.MEDIAN_COVERAGE': 'qc.median_coverage',
            },
            qc_meta_map={
                'raw_data.FREEMIX': 'freemix',
                'raw_data.PCT_CHIMERAS': 'pct_chimeras',
                'raw_data.MEDIAN_INSERT_SIZE': 'median_insert_size',
                'raw_data.MEDIAN_COVERAGE': 'median_coverage',
            },
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
            reads_column='CRAM',
            gvcf_column='GVCF',
        )

        parser.filename_map = {
            '<sample-id>.g.vcf.gz': '/path/to/<sample-id>.g.vcf.gz',
            '<sample-id>.bam': '/path/to/<sample-id>.bam',
        }

        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(1, len(resp['samples']['insert']))
        self.assertEqual(1, len(resp['sequences']['insert']))
        self.assertEqual(0, len(resp['samples']['update']))
        self.assertEqual(0, len(resp['sequences']['update']))
        self.assertEqual(1, len(sum(resp['analyses'].values(), [])))

        samples_to_add = resp['samples']['insert']
        sequences_to_add = resp['sequences']['insert']
        analyses_to_add = resp['analyses']

        self.assertDictEqual({'centre': 'KCCG'}, samples_to_add[0].meta)
        expected_sequence_dict = {
            'qc': {
                'median_insert_size': '400',
                'median_coverage': '30',
                'freemix': '0.01',
                'pct_chimeras': '0.01',
            },
            'reads': [
                {
                    'location': '/path/to/<sample-id>.bam',
                    'basename': '<sample-id>.bam',
                    'class': 'File',
                    'checksum': None,
                    'size': 111,
                }
            ],
            'reads_type': 'bam',
            'gvcfs': [
                {
                    'location': '/path/to/<sample-id>.g.vcf.gz',
                    'basename': '<sample-id>.g.vcf.gz',
                    'class': 'File',
                    'checksum': None,
                    'size': 111,
                }
            ],
            'gvcf_types': 'gvcf',
        }
        self.assertDictEqual(expected_sequence_dict, sequences_to_add[0].meta)
        analysis = analyses_to_add['<sample-id>'][0]
        self.assertDictEqual(
            {
                'median_insert_size': '400',
                'median_coverage': '30',
                'freemix': '0.01',
                'pct_chimeras': '0.01',
            },
            analysis.meta,
        )

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    async def test_rows_with_participants(
        self,
        mock_get_sequence_ids,
        mock_get_sample_id,
        mock_get_participant_id_map_by_external_ids,
    ):
        """
        Test importing a single row with a participant id, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_sequence_ids_for_sample_ids_by_type
        """
        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}
        mock_get_participant_id_map_by_external_ids.return_value = {}

        rows = [
            'Individual ID\tSample ID\tFilenames\tType',
            'Demeter\tsample_id001\tsample_id001.filename-R1.fastq.gz,sample_id001.filename-R2.fastq.gz\tWGS',
            'Demeter\tsample_id001\tsample_id001.exome.filename-R1.fastq.gz,sample_id001.exome.filename-R2.fastq.gz\tWES',
            'Apollo\tsample_id002\tsample_id002.filename-R1.fastq.gz\tWGS',
            'Apollo\tsample_id002\tsample_id002.filename-R2.fastq.gz\tWGS',
            'Athena\tsample_id003\tsample_id003.filename-R1.fastq.gz',
            'Athena\tsample_id003\tsample_id003.filename-R2.fastq.gz',
            'Apollo\tsample_id004\tsample_id004.filename-R1.fastq.gz',
            'Apollo\tsample_id004\tsample_id004.filename-R2.fastq.gz',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_column='Individual ID',
            sample_name_column='Sample ID',
            reads_column='Filenames',
            seq_type_column='Type',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
        )

        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'sample_id001.filename-R1.fastq.gz': '/path/to/sample_id001.filename-R1.fastq.gz',
            'sample_id001.filename-R2.fastq.gz': '/path/to/sample_id001.filename-R2.fastq.gz',
            'sample_id001.exome.filename-R1.fastq.gz': '/path/to/sample_id001.exome.filename-R1.fastq.gz',
            'sample_id001.exome.filename-R2.fastq.gz': '/path/to/sample_id001.exome.filename-R2.fastq.gz',
            'sample_id002.filename-R1.fastq.gz': '/path/to/sample_id002.filename-R1.fastq.gz',
            'sample_id002.filename-R2.fastq.gz': '/path/to/sample_id002.filename-R2.fastq.gz',
            'sample_id003.filename-R1.fastq.gz': '/path/to/sample_id003.filename-R1.fastq.gz',
            'sample_id003.filename-R2.fastq.gz': '/path/to/sample_id003.filename-R2.fastq.gz',
            'sample_id004.filename-R1.fastq.gz': '/path/to/sample_id004.filename-R1.fastq.gz',
            'sample_id004.filename-R2.fastq.gz': '/path/to/sample_id004.filename-R2.fastq.gz',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(3, len(resp['participants']['insert']))
        self.assertEqual(0, len(resp['participants']['update']))
        self.assertEqual(4, len(resp['samples']['insert']))
        self.assertEqual(0, len(resp['samples']['update']))
        self.assertEqual(5, len(resp['sequences']['insert']))
        self.assertEqual(0, len(resp['sequences']['update']))
        self.assertEqual(0, len(sum(resp['analyses'].values(), [])))

        participants_to_add = resp['participants']['insert']
        sequences_to_add = resp['sequences']['insert']

        expected_sequence_dict = {
            'reads': [
                [
                    {
                        'basename': 'sample_id001.filename-R1.fastq.gz',
                        'checksum': None,
                        'class': 'File',
                        'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                        'size': None,
                    },
                    {
                        'basename': 'sample_id001.filename-R2.fastq.gz',
                        'checksum': None,
                        'class': 'File',
                        'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                        'size': None,
                    },
                ]
            ],
            'reads_type': 'fastq',
        }
        self.assertDictEqual(expected_sequence_dict, sequences_to_add[0].meta)

        # Check that both of Demeter's sequences are there
        self.assertEqual(participants_to_add[0].external_id, 'Demeter')
        self.assertEqual(len(participants_to_add[0].samples), 1)
        self.assertEqual(len(participants_to_add[0].samples[0].sequences), 2)

        return

    @run_as_sync
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    async def test_rows_with_valid_participant_meta(
        self,
        mock_get_sequence_ids,
        mock_get_sample_id,
        mock_get_participant_id_map_by_external_ids,
    ):
        """
        Test importing a several rows with a participant metadata (reported gender, sex and karyotype),
        forms objects and checks response
        - MOCKS: get_sample_id_map_by_external,  get_participant_id_map_by_external_ids,
        get_sequence_ids_for_sample_ids_by_type
        """

        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}
        mock_get_participant_id_map_by_external_ids.return_value = {}

        rows = [
            'Individual ID\tSample ID\tSex\tGender\tKaryotype',
            'Demeter\tsample_id001\tMale\tNon-binary\tXY',
            'Apollo\tsample_id002\tFemale\tFemale\tXX',
            'Athena\tsample_id003\tFEMalE',
            'Dionysus\tsample_id00x\t\tMale\tXX',
            'Pluto\tsample_id00y',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_column='Individual ID',
            sample_name_column='Sample ID',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            reported_sex_column='Sex',
            reported_gender_column='Gender',
            karyotype_column='Karyotype',
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
        )

        # Call generic parser
        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )
        participants_to_add = resp['participants']['insert']

        # Assert that the participant meta is there.
        self.assertEqual(participants_to_add[0].reported_gender, 'Non-binary')
        self.assertEqual(participants_to_add[0].reported_sex, 1)
        self.assertEqual(participants_to_add[0].karyotype, 'XY')
        self.assertEqual(participants_to_add[1].reported_gender, 'Female')
        self.assertEqual(participants_to_add[1].reported_sex, 2)
        self.assertEqual(participants_to_add[1].karyotype, 'XX')
        self.assertEqual(participants_to_add[2].reported_sex, 2)
        self.assertEqual(participants_to_add[2].get('reported_gender'), None)
        self.assertEqual(participants_to_add[2].get('karyotype'), None)
        self.assertEqual(participants_to_add[3].reported_gender, 'Male')
        self.assertEqual(participants_to_add[3].karyotype, 'XX')
        return

    @run_as_sync
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    async def test_rows_with_invalid_participant_meta(
        self,
        mock_get_sequence_ids,
        mock_get_sample_id,
        mock_get_participant_id_map_by_external_ids,
    ):
        """
        Test importing a single rows with invalid participant metadata,
        forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_participant_id_map_by_external_ids
        """

        mock_get_sequence_ids.return_value = {}
        mock_get_sample_id.return_value = {}
        mock_get_participant_id_map_by_external_ids.return_value = {}

        rows = [
            'Individual ID\tSample ID\tSex\tKaryotype',
            'Athena\tsample_id003\tFemalee\tXX',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            participant_column='Individual ID',
            sample_name_column='Sample ID',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            reported_sex_column='Sex',
            karyotype_column='Karyotype',
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
        )

        # Call generic parser
        file_contents = '\n'.join(rows)
        with self.assertRaises(ValueError):
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        return

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.parser.cloudhelper.AnyPath')
    async def test_cram_with_no_reference(
        self,
        mock_any_path,
        mock_get_sequence_ids,
        mock_get_sample_id,
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_get_sequence_ids.return_value = {}
        mock_get_sample_id.return_value = {}

        mock_any_path.return_value.stat.return_value.st_size = 111
        mock_any_path.return_value.exists.return_value = False

        rows = [
            'Sample ID\tFilename',
            'sample_id003\tfile.cram',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_name_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
        )

        parser.filename_map = {'file.cram': 'gs://path/file.cram'}

        # Call generic parser
        file_contents = '\n'.join(rows)
        with self.assertRaises(ValueError) as ctx:
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        self.assertEqual(
            'Reads type for "sample_id003" is CRAM, but a reference is not defined, please set the default reference assembly path',
            str(ctx.exception),
        )

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.parser.cloudhelper.AnyPath')
    async def test_cram_with_default_reference(
        self,
        mock_any_path,
        mock_get_sequence_ids,
        mock_get_sample_id,
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_get_sequence_ids.return_value = {}
        mock_get_sample_id.return_value = {}

        mock_any_path.return_value.stat.return_value.st_size = 111
        mock_any_path.return_value.exists.return_value = True

        rows = [
            'Sample ID\tFilename',
            'sample_id003\tfile.cram',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_name_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
            default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file.fasta': 'gs://path/file.fasta',
            'file.fasta.fai': 'gs://path/file.fasta.fai',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        expected = {
            'location': 'gs://path/file.fasta',
            'basename': 'file.fasta',
            'class': 'File',
            'checksum': None,
            'size': None,
            'secondaryFiles': [
                {
                    'location': 'gs://path/file.fasta.fai',
                    'basename': 'file.fasta.fai',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                }
            ],
        }

        self.assertDictEqual(
            expected,
            resp['sequences']['insert'][0]['meta']['reference_assembly'],
        )

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.parser.cloudhelper.AnyPath')
    async def test_cram_with_row_level_reference(
        self,
        mock_any_path,
        mock_get_sequence_ids,
        mock_get_sample_id,
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_get_sequence_ids.return_value = {}
        mock_get_sample_id.return_value = {}

        mock_any_path.return_value.exists.return_value = True

        rows = [
            'Sample ID\tFilename\tRef',
            'sample_id003\tfile.cram\tref.fa',
            'sample_id003\tfile2.cram\tref.fa',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_name_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
            reference_assembly_location_column='Ref'
            # default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file2.cram': 'gs://path/file2.cram',
            'ref.fa': 'gs://path/ref.fa',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        expected = {
            'location': 'gs://path/ref.fa',
            'basename': 'ref.fa',
            'class': 'File',
            'checksum': None,
            'size': None,
            'secondaryFiles': [
                {
                    'location': 'gs://path/ref.fa.fai',
                    'basename': 'ref.fa.fai',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                }
            ],
        }

        self.assertDictEqual(
            expected,
            resp['sequences']['insert'][0]['meta']['reference_assembly'],
        )

    @run_as_sync
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    @patch('sample_metadata.parser.cloudhelper.AnyPath')
    async def test_cram_with_multiple_row_level_references(
        self,
        mock_any_path,
        mock_get_sequence_ids,
        mock_get_sample_id,
    ):
        """
        Test importing a single row with a cram with no reference
        This should throw an exception
        """

        mock_get_sequence_ids.return_value = {}
        mock_get_sample_id.return_value = {}

        mock_any_path.return_value.exists.return_value = True

        rows = [
            'Sample ID\tFilename\tRef',
            'sample_id003\tfile.cram\tref.fa',
            'sample_id003\tfile2.cram\tref2.fa',
        ]

        parser = GenericMetadataParser(
            search_locations=[],
            sample_name_column='Sample ID',
            reads_column='Filename',
            participant_meta_map={},
            sample_meta_map={},
            sequence_meta_map={},
            qc_meta_map={},
            # doesn't matter, we're going to mock the call anyway
            project='devdev',
            reference_assembly_location_column='Ref'
            # default_reference_assembly_location='gs://path/file.fasta',
        )
        parser.skip_checking_gcs_objects = True
        parser.filename_map = {
            'file.cram': 'gs://path/file.cram',
            'file2.cram': 'gs://path/file.cram',
            'ref.fa': 'gs://path/ref.fa',
            'ref2.fa': 'gs://path/ref2.fa',
        }

        # Call generic parser
        file_contents = '\n'.join(rows)

        with self.assertRaises(ValueError) as ctx:
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        self.assertEqual(
            'Multiple reference assemblies were defined for sample_id003: ref.fa, ref2.fa',
            str(ctx.exception),
        )



# pylint: disable=invalid-overridden-method
from datetime import date, timedelta

from test.testbase import DbIsolatedTest, run_as_sync

from models.enums.sequencing import SequenceType
from models.enums import AnalysisType, AnalysisStatus

from db.python.layers.analysis import AnalysisLayer
from db.python.layers.sample import SampleLayer, SampleType


class TestAnalysis(DbIsolatedTest):
    """Test sample class"""

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()
        self.sl = SampleLayer(self.connection)
        self.al = AnalysisLayer(self.connection)

        self.sample_id = await self.sl.insert_sample(
            'Test01',
            SampleType.BLOOD,
            meta={'meta': 'meta ;)'},
            active=True,
        )

        await self.al.insert_analysis(
            analysis_type=AnalysisType.CRAM,
            status=AnalysisStatus.COMPLETED,
            sample_ids=[self.sample_id],
            meta={'sequence_type': 'genome', 'size': 1024},
        )

    @run_as_sync
    async def test_add_cram(self):
        """
        Test adding an analysis of type CRAM
        """

        analyses = await self.connection.connection.fetch_all('SELECT * FROM analysis')
        analysis_samples = await self.connection.connection.fetch_all(
            'SELECT * FROM analysis_sample'
        )
        print(analyses, analysis_samples)

        self.assertEqual(1, len(analyses))

        self.assertEqual(1, analysis_samples[0]['sample_id'])
        self.assertEqual(analyses[0]['id'], analysis_samples[0]['analysis_id'])

    @run_as_sync
    async def test_get_sample_file_sizes(self):
        """
        Test retrieval of sample file sizes over time
        """

        result = await self.al.get_sample_file_sizes(project_ids=[1])
        expected = [
            {
                'project': 1,
                'samples': [
                    {
                        'sample': self.sample_id,
                        'dates': [
                            {
                                'start': date.today(),
                                'end': None,
                                'size': {SequenceType.GENOME: 1024},
                            }
                        ],
                    }
                ],
            }
        ]

        # Check output is formatted correctly
        self.assertEqual(1, len(result))
        self.assertDictEqual(expected[0], result[0])

        # Add exome cram
        await self.al.insert_analysis(
            analysis_type=AnalysisType.CRAM,
            status=AnalysisStatus.COMPLETED,
            sample_ids=[self.sample_id],
            meta={'sequence_type': 'exome', 'size': 3141},
        )

        expected[0]['samples'][0]['dates'][0]['size'][SequenceType.EXOME] = 3141

        # Assert that the exome size was added correctly
        result = await self.al.get_sample_file_sizes(project_ids=[1])
        self.assertDictEqual(expected[0], result[0])

        # Assert that if we select a date range outside of sample creation date
        # that is doesn't show up in the map
        yesterday = date.today() - timedelta(days=1)
        result = await self.al.get_sample_file_sizes(
            project_ids=[1], end_date=yesterday
        )

        self.assertEqual([], result)

        # Add another genome cram that's newer
        await self.al.insert_analysis(
            analysis_type=AnalysisType.CRAM,
            status=AnalysisStatus.COMPLETED,
            sample_ids=[self.sample_id],
            meta={'sequence_type': 'genome', 'size': 11111},
        )

        expected[0]['samples'][0]['dates'][0]['size'][SequenceType.GENOME] = 11111
        result = await self.al.get_sample_file_sizes(project_ids=[1])
        self.assertDictEqual(expected[0], result[0])

        # Add another sample and it's analysis cram as well
        sample_id_2 = await self.sl.insert_sample(
            'Test02',
            SampleType.BLOOD,
            meta={'meta': 'meta ;)'},
            active=True,
        )
        await self.al.insert_analysis(
            analysis_type=AnalysisType.CRAM,
            status=AnalysisStatus.COMPLETED,
            sample_ids=[sample_id_2],
            meta={'sequence_type': 'genome', 'size': 987654321},
        )

        sample_2_data = {
            'sample': sample_id_2,
            'dates': [
                {
                    'start': date.today(),
                    'end': None,
                    'size': {SequenceType.GENOME: 987654321},
                }
            ],
        }
        expected[0]['samples'].append(sample_2_data)

        result = await self.al.get_sample_file_sizes(project_ids=[1])
        self.assertDictEqual(expected[0], result[0])



import unittest
from io import StringIO
from unittest.mock import patch

from test.testbase import run_as_sync

from scripts.parse_ont_sheet import OntParser


class TestOntSampleSheetParser(unittest.TestCase):
    """Test the TestOntSampleSheetParser"""

    @run_as_sync
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequences_by_sample_ids')
    async def test_simple_sheet(
        self, mock_get_sequence_ids, mock_get_sample_id, mock_get_participant_id
    ):
        """
        Test importing a two rows, forms objects and checks response
        - MOCKS:
            - get_participant_id_map_by_external_ids
            - get_sample_id_map_by_external
            - get_sequence_ids_for_sample_ids_by_type
        """
        mock_get_participant_id.return_value = {'Sample01': 1}
        mock_get_sample_id.return_value = {'Sample01': 'CPG001'}
        mock_get_sequence_ids.return_value = {}

        rows = [
            'Sequencing_date,Experiment name,Sample ID,Protocol,Flow cell,Barcoding,Device,Flowcell ID,MUX total,Basecalling,Fail FASTQ filename,Pass FASTQ filename',
            '10/12/2034,PBXP_Awesome,Sample01,LSK1,PRO002,None,PromethION,XYZ1,7107,4.0.11+f1071ce,Sample01_fail.fastq.gz,Sample01_pass.fastq.gz',
            '21/10/2021,PBXP_Awesome,Sample02,LSK1,PRO002,None,PromethION,XYZ1,8056,4.0.11+f1071ce,Sample02_fail.fastq.gz,Sample02_pass.fastq.gz',
        ]
        parser = OntParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project='dev',
        )

        parser.skip_checking_gcs_objects = True
        fs = [
            'Sample01_pass.fastq.gz',
            'Sample01_fail.fastq.gz',
            'Sample02_pass.fastq.gz',
            'Sample02_fail.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter=',', dry_run=True
        )

        participants_to_add = resp['participants']['insert']
        participants_to_update = resp['participants']['update']
        samples_to_add = resp['samples']['insert']
        samples_to_update = resp['samples']['update']
        sequencing_to_add = resp['sequences']['insert']
        sequencing_to_update = resp['sequences']['update']

        self.assertEqual(1, len(participants_to_add))
        self.assertEqual(1, len(participants_to_update))
        self.assertEqual(1, len(samples_to_add))
        self.assertEqual(2, len(sequencing_to_add))
        self.assertEqual(1, len(samples_to_update))
        self.assertEqual(0, len(sequencing_to_update))

        meta_dict = {
            'barcoding': 'None',
            'basecalling': '4.0.11+f1071ce',
            'device': 'PromethION',
            'experiment_name': 'PBXP_Awesome',
            'failed_reads': [
                [
                    {
                        'location': 'gs://BUCKET/FAKE/Sample01_fail.fastq.gz',
                        'basename': 'Sample01_fail.fastq.gz',
                        'class': 'File',
                        'checksum': None,
                        'size': None,
                    }
                ]
            ],
            'flow_cell': 'PRO002',
            'flowcell_id': 'XYZ1',
            'mux_total': '7107',
            'protocol': 'LSK1',
            'reads': [
                [
                    {
                        'basename': 'Sample01_pass.fastq.gz',
                        'checksum': None,
                        'class': 'File',
                        'location': 'gs://BUCKET/FAKE/Sample01_pass.fastq.gz',
                        'size': None,
                    }
                ]
            ],
            'reads_type': 'fastq',
            'sequencing_date': '10/12/2034',
        }

        self.assertDictEqual(meta_dict, sequencing_to_add[0].meta)



from test.testbase import DbIsolatedTest, run_as_sync

from models.models.sample import sample_id_format
from db.python.layers.sample import SampleLayer, SampleType


class TestSample(DbIsolatedTest):
    """Test sample class"""

    # tests run in 'sorted by ascii' order
    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.slayer = SampleLayer(self.connection)

    @run_as_sync
    async def test_add_sample(self):
        """Test inserting a sample"""
        s = await self.slayer.insert_sample(
            'Test01',
            SampleType.BLOOD,
            active=True,
            meta={'meta': 'meta ;)'},
        )

        samples = await self.connection.connection.fetch_all(
            'SELECT id, type, meta, project FROM sample'
        )
        self.assertEqual(1, len(samples))
        s = samples[0]
        self.assertEqual(1, s['id'])

    @run_as_sync
    async def test_get_sample(self):
        """Test getting formed sample"""
        meta_dict = {'meta': 'meta ;)'}
        s = await self.slayer.insert_sample(
            'Test01',
            SampleType.BLOOD,
            active=True,
            meta=meta_dict,
        )

        sample = await self.slayer.get_by_id(s, check_project_id=False)

        self.assertEqual(sample_id_format(s), sample.id)
        self.assertEqual(SampleType.BLOOD, sample.type)
        self.assertDictEqual(meta_dict, sample.meta)

    # @run_as_sync
    # async def test_update_sample(self):
    #     """Test to see what's in the database"""
    #     print('Running test 2')
    #     print(await self.connection.connection.fetch_all('SELECT * FROM sample'))



import unittest
from io import StringIO
from unittest.mock import patch

from test.testbase import run_as_sync

from sample_metadata.parser.sample_file_map_parser import SampleFileMapParser


class TestSampleMapParser(unittest.TestCase):
    """Test the TestSampleMapParser"""

    @run_as_sync
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    async def test_single_row_fastq(
        self, mock_get_sequence_ids, mock_get_sample_id, mock_participant_ids
    ):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_sequence_ids_for_sample_ids_by_type
        """
        mock_participant_ids.return_value = {}
        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}

        rows = [
            'Individual ID\tFilenames',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz',
        ]
        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project='dev',
        )
        fs = ['<sample-id>.filename-R1.fastq.gz', '<sample-id>.filename-R2.fastq.gz']
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        participants_to_add = resp['participants']['insert']
        participants_to_update = resp['participants']['update']
        samples_to_add = resp['samples']['insert']
        samples_to_update = resp['samples']['update']
        sequencing_to_add = resp['sequences']['insert']
        sequencing_to_update = resp['sequences']['update']

        self.assertEqual(1, len(participants_to_add))
        self.assertEqual(0, len(participants_to_update))
        self.assertEqual(1, len(samples_to_add))
        self.assertEqual(0, len(samples_to_update))
        self.assertEqual(1, len(sequencing_to_add))
        self.assertEqual(0, len(sequencing_to_update))

        self.assertDictEqual({}, samples_to_add[0].meta)
        expected_sequence_dict = {
            'reads': [
                [
                    {
                        'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R1.fastq.gz',
                        'basename': '<sample-id>.filename-R1.fastq.gz',
                        'class': 'File',
                        'checksum': None,
                        'size': None,
                    },
                    {
                        'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R2.fastq.gz',
                        'basename': '<sample-id>.filename-R2.fastq.gz',
                        'class': 'File',
                        'checksum': None,
                        'size': None,
                    },
                ]
            ],
            'reads_type': 'fastq',
        }
        self.assertDictEqual(expected_sequence_dict, sequencing_to_add[0].meta)

    @run_as_sync
    @patch('sample_metadata.apis.ParticipantApi.get_participant_id_map_by_external_ids')
    @patch('sample_metadata.apis.SampleApi.get_sample_id_map_by_external')
    @patch('sample_metadata.apis.SequenceApi.get_sequence_ids_for_sample_ids_by_type')
    async def test_two_rows_with_provided_checksums(
        self, mock_get_sequence_ids, mock_get_sample_id, mock_participant_ids
    ):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_sequence_ids_for_sample_ids_by_type
        """
        mock_participant_ids.return_value = {}
        mock_get_sample_id.return_value = {}
        mock_get_sequence_ids.return_value = {}

        rows = [
            'Individual ID\tFilenames\tChecksum',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz\t<checksum>,<checksum2>',
            '<sample-id2>\t<sample-id2>.filename-R1.fastq.gz\t<checksum3>',
            '<sample-id2>\t<sample-id2>.filename-R2.fastq.gz\t<checksum4>',
        ]
        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project='dev',
        )
        fs = [
            '<sample-id>.filename-R1.fastq.gz',
            '<sample-id>.filename-R2.fastq.gz',
            '<sample-id2>.filename-R1.fastq.gz',
            '<sample-id2>.filename-R2.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        resp = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        participants_to_add = resp['participants']['insert']
        participants_to_update = resp['participants']['update']
        samples_to_add = resp['samples']['insert']
        samples_to_update = resp['samples']['update']
        sequencing_to_add = resp['sequences']['insert']
        sequencing_to_update = resp['sequences']['update']

        self.assertEqual(2, len(participants_to_add))
        self.assertEqual(0, len(participants_to_update))
        self.assertEqual(2, len(samples_to_add))
        self.assertEqual(0, len(samples_to_update))
        self.assertEqual(2, len(sequencing_to_add))
        self.assertEqual(0, len(sequencing_to_update))

        self.assertDictEqual({}, samples_to_add[0].meta)
        expected_sequence1_reads = [
            {
                'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R1.fastq.gz',
                'basename': '<sample-id>.filename-R1.fastq.gz',
                'class': 'File',
                'checksum': '<checksum>',
                'size': None,
            },
            {
                'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R2.fastq.gz',
                'basename': '<sample-id>.filename-R2.fastq.gz',
                'class': 'File',
                'checksum': '<checksum2>',
                'size': None,
            },
        ]

        self.assertListEqual(
            expected_sequence1_reads, sequencing_to_add[0].meta['reads'][0]
        )

        expected_sequence2_reads = [
            {
                'location': 'gs://BUCKET/FAKE/<sample-id2>.filename-R1.fastq.gz',
                'basename': '<sample-id2>.filename-R1.fastq.gz',
                'class': 'File',
                'checksum': '<checksum3>',
                'size': None,
            },
            {
                'location': 'gs://BUCKET/FAKE/<sample-id2>.filename-R2.fastq.gz',
                'basename': '<sample-id2>.filename-R2.fastq.gz',
                'class': 'File',
                'checksum': '<checksum4>',
                'size': None,
            },
        ]
        self.assertListEqual(
            expected_sequence2_reads, sequencing_to_add[1].meta['reads'][0]
        )



# pylint: disable=too-many-lines,too-many-instance-attributes,too-many-locals,unused-argument,assignment-from-none,invalid-name,ungrouped-imports
import sys
import asyncio
import csv
import logging
import os
import re
from abc import abstractmethod
from collections import defaultdict
from typing import (
    List,
    Dict,
    Union,
    Optional,
    Tuple,
    Match,
    Any,
    Sequence,
    TypeVar,
    Iterator,
    Coroutine,
    Set,
    Iterable,
)
from functools import wraps

from cloudpathlib import AnyPath

from api.utils import group_by

from sample_metadata.parser.cloudhelper import CloudHelper

from sample_metadata.model.participant_upsert_body import ParticipantUpsertBody

from sample_metadata.apis import SampleApi, SequenceApi, AnalysisApi, ParticipantApi
from sample_metadata.models import (
    ParticipantUpsert,
    SequenceType,
    AnalysisModel,
    SampleType,
    AnalysisType,
    SequenceStatus,
    AnalysisStatus,
    SampleBatchUpsert,
    SampleBatchUpsertBody,
    SequenceUpsert,
)


# https://mypy.readthedocs.io/en/stable/runtime_troubles.html#using-new-additions-to-the-typing-module
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

logging.basicConfig()
logger = logging.getLogger(__file__)
logger.setLevel(logging.INFO)

FASTQ_EXTENSIONS = ('.fq', '.fastq', '.fq.gz', '.fastq.gz')
BAM_EXTENSIONS = ('.bam',)
CRAM_EXTENSIONS = ('.cram',)
GVCF_EXTENSIONS = ('.g.vcf.gz',)
VCF_EXTENSIONS = ('.vcf', '.vcf.gz')

ALL_EXTENSIONS = (
    FASTQ_EXTENSIONS
    + CRAM_EXTENSIONS
    + BAM_EXTENSIONS
    + GVCF_EXTENSIONS
    + VCF_EXTENSIONS
)

# construct rmatch string to capture all fastq patterns
rmatch_str = (
    r'[_\.-][Rr]?[12]('
    + '|'.join(s.replace('.', '\\.') for s in FASTQ_EXTENSIONS)
    + ')$'
)
rmatch = re.compile(rmatch_str)
SingleRow = Dict[str, Any]
GroupedRow = List[SingleRow]

T = TypeVar('T')

SUPPORTED_FILE_TYPE = Literal['reads', 'variants']
SUPPORTED_READ_TYPES = Literal['fastq', 'bam', 'cram']
SUPPORTED_VARIANT_TYPES = Literal['gvcf', 'vcf']


class CustomDictReader(csv.DictReader):
    """csv.DictReader that strips whitespace off headers"""

    def __init__(
        self,
        *args,
        key_map=None,
        required_keys: Iterable[str] = None,
        ignore_extra_keys=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.key_map = key_map
        self.ignore_extra_keys = ignore_extra_keys
        self.required_keys = set(required_keys) if required_keys else None
        self._custom_cached_fieldnames = None

    @property
    def fieldnames(self):
        if not self._custom_cached_fieldnames:
            fs = super().fieldnames
            self._custom_cached_fieldnames = list(map(self.process_fieldname, fs))

            if self.required_keys:
                missing_keys = self.required_keys - set(self._custom_cached_fieldnames)
                if missing_keys:
                    raise ValueError('Missing keys: ' + ','.join(missing_keys))
        return self._custom_cached_fieldnames

    def process_fieldname(self, fieldname: str) -> str:
        """
        Process the fieldname
        (default: strip leading / trailing whitespace)
        """
        if self.key_map:
            for k, values in self.key_map.items():
                if fieldname.lower() in values:
                    return k

            if not self.ignore_extra_keys:
                raise ValueError(
                    f'Key "{fieldname}" not found in provided key map: {", ".join(self.key_map.keys())}'
                )

        return fieldname.strip()


class ParticipantMetaGroup:
    """Class for holding participant metadata grouped by id"""

    def __init__(
        self,
        participant_id: int,
        rows: GroupedRow,
        meta: Dict[str, Any],
    ):
        self.id = participant_id
        self.rows = rows
        self.meta = meta


class SampleMetaGroup:
    """Class for holding sample metadata grouped by id"""

    def __init__(
        self,
        sample_id: str,
        rows: GroupedRow,
        meta: Dict[str, Any] = None,
    ):
        self.id = sample_id
        self.rows = rows
        self.meta = meta


class SequenceMetaGroup:
    """Class for holding sequence metadata grouped by type"""

    def __init__(
        self,
        rows: GroupedRow,
        sequence_type: SequenceType,
        meta: Optional[Dict[str, Any]] = None,
    ):
        self.rows = rows
        self.sequence_type = sequence_type
        self.meta = meta


def chunk(iterable: Iterable[T], chunk_size=50) -> Iterator[List[T]]:
    """
    Chunk a sequence by yielding lists of `chunk_size`
    """
    chnk: List[T] = []
    for element in iterable:
        chnk.append(element)
        if len(chnk) >= chunk_size:
            yield chnk
            chnk = []

    if chnk:
        yield chnk


def run_as_sync(f):
    """
    Run an async function, synchronously.
    Useful for @click functions that must be async, eg:

    @click.command()
    @click.option(...)
    @run_as_sync
    async def my_async_function(**kwargs):
        return await awaitable_function(**kwargs)
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


class GenericParser(
    CloudHelper
):  # pylint: disable=too-many-public-methods,too-many-arguments
    """Parser for VCGS manifest"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        path_prefix: Optional[str],
        search_paths: list[str],
        project: str,
        default_sequence_type='genome',
        default_sequence_status='uploaded',
        default_sample_type='blood',
        default_analysis_type='qc',
        default_analysis_status='completed',
        skip_checking_gcs_objects=False,
        key_map: Dict[str, str] = None,
        ignore_extra_keys=False,
        required_keys: Set[str] = None,
        verbose=True,
    ):

        self.path_prefix = path_prefix
        self.skip_checking_gcs_objects = skip_checking_gcs_objects
        self.verbose = verbose

        self.key_map = key_map
        self.required_keys = required_keys
        self.ignore_extra_keys = ignore_extra_keys

        if not project:
            raise ValueError('sample-metadata project is required')

        self.project = project

        self.default_sequence_type: str = default_sequence_type
        self.default_sequence_status: str = default_sequence_status
        self.default_sample_type: str = default_sample_type
        self.default_analysis_type: str = default_analysis_type
        self.default_analysis_status: str = default_analysis_status

        # gs specific
        self.default_bucket = None

        self._client = None
        self.bucket_clients: Dict[str, Any] = {}

        self.papi = ParticipantApi()
        self.sapi = SampleApi()
        self.seqapi = SequenceApi()

        super().__init__(search_paths)

    def file_path(self, filename: str, raise_exception: bool = True) -> str | None:
        """
        Get complete filepath of filename:
        - Includes gs://{bucket} if relevant
        - Includes path_prefix decided early on
        """
        if filename.startswith('gs://') or filename.startswith('/'):
            return filename

        if not self.path_prefix:
            fn = super().file_path(filename, raise_exception=raise_exception)
            if not fn:
                raise FileNotFoundError(
                    f'Cannot form full path to "{filename}" as '
                    'no path_prefix was defined'
                )
            return fn

        return os.path.join(self.path_prefix or '', filename)

    @abstractmethod
    def get_sample_id(self, row: SingleRow) -> Optional[str]:
        """Get external sample ID from row"""

    # @abstractmethod
    def get_sequence_id(self, row: GroupedRow) -> Optional[dict[str, str]]:
        """Get external sequence ID from row"""
        return None

    @abstractmethod
    def get_participant_id(self, row: SingleRow) -> Optional[str]:
        """Get external participant ID from row"""

    def get_reported_sex(self, row: GroupedRow) -> Optional[int]:
        """Get reported sex from grouped row"""
        return None

    def get_reported_gender(self, row: GroupedRow) -> Optional[str]:
        """Get reported gender from grouped row"""
        return None

    def get_karyotype(self, row: GroupedRow) -> Optional[str]:
        """Get karyotype from grouped row"""
        return None

    # @abstractmethod
    def has_participants(self, file_pointer, delimiter: str) -> bool:
        """Returns True if the file has a Participants column"""

    # @abstractmethod
    async def get_grouped_sample_meta(self, rows: GroupedRow) -> List[SampleMetaGroup]:
        """Return list of grouped by sample metadata from the rows"""

    @abstractmethod
    async def get_sample_meta(self, sample_group: SampleMetaGroup) -> SampleMetaGroup:
        """Get sample-metadata from row"""

    # @abstractmethod
    async def get_grouped_sequence_meta(
        self, sample_id: str, rows: GroupedRow
    ) -> List[SequenceMetaGroup]:
        """Return list of grouped by type sequence metadata from the rows"""

    @abstractmethod
    async def get_sequence_meta(
        self, seq_group: SequenceMetaGroup
    ) -> SequenceMetaGroup:
        """Get sequence-metadata from row then set it in the SequenceMetaGroup"""
        return SequenceMetaGroup(
            rows=[], sequence_type=self.default_sequence_type, meta={}
        )

    # @abstractmethod
    async def get_participant_meta(
        self, participant_id: int, rows: GroupedRow
    ) -> ParticipantMetaGroup:
        """Get participant-metadata from rows then set it in the ParticipantMetaGroup"""
        return ParticipantMetaGroup(participant_id=participant_id, rows=rows, meta={})

    # @abstractmethod
    async def get_analyses(
        self, sample_id: str, row: SingleRow, cpg_id: Optional[str]
    ) -> List[AnalysisModel]:
        """
        Get analysis objects from row. Optionally, a CPG ID can be passed for
        to help finding files for previously added samples that were already renamed.
        """
        return []

    @abstractmethod
    async def get_qc_meta(self, sample_id: str, row: SingleRow) -> Optional[SingleRow]:
        """Get qc-meta from row, creates a Analysis object of type QC"""

    # @abstractmethod
    def get_sample_type(self, row: GroupedRow) -> SampleType:
        """Get sample type from row"""
        return SampleType(self.default_sample_type)

    # @abstractmethod
    def get_sequence_types(self, row: GroupedRow) -> List[SequenceType]:
        """Get sequence types from row"""
        return [SequenceType(self.default_sequence_type)]

    # @abstractmethod
    def get_sequence_type(self, row: SingleRow) -> SequenceType:
        """Get sequence types from row"""
        return SequenceType(self.default_sequence_type)

    # @abstractmethod
    def get_sequence_status(self, row: GroupedRow) -> SequenceStatus:
        """Get sequence status from row"""
        return SequenceStatus(self.default_sequence_status)

    # @abstractmethod
    def get_analysis_type(self, sample_id: str, row: GroupedRow) -> AnalysisType:
        """Get analysis type from row"""
        return AnalysisType(self.default_analysis_type)

    # @abstractmethod
    def get_analysis_status(self, sample_id: str, row: GroupedRow) -> AnalysisStatus:
        """Get analysis status from row"""
        return AnalysisStatus(self.default_analysis_status)

    @staticmethod
    def get_existing_external_sequence_ids(
        participant_map: Dict[str, Dict[Any, List[Any]]]
    ):
        """Pulls external sequence IDs from participant map"""
        external_sequence_ids: list[str] = []
        for participant in participant_map:
            for sample in participant_map[participant]:
                for sequence in participant_map[participant][sample]:
                    external_sequence_ids.append((sequence.get('Sequence ID')))

        return external_sequence_ids

    @staticmethod
    def get_existing_sequences(
        sequences: list[dict[str, Any]], external_sequence_ids: list[str]
    ):
        """Accounts for external_sequence_ids when determining which sequences
        need to be updated vs inserted"""

        existing_sequences: list[dict[str, Any]] = []
        for seq in sequences:
            if not seq['external_ids'].values():
                # No existing sequence ID, we can assume that replacement should happen
                # Note: This means that you can't have a mix of sequences with and without
                # external sequence IDs in one dataset.
                existing_sequences.append(seq)

            else:
                for ext_id in seq['external_ids'].values():
                    # If the external ID is already there, we want to upsert.
                    if ext_id in external_sequence_ids:
                        existing_sequences.append(seq)

        return existing_sequences

    async def process_sample_group(
        self,
        rows: GroupedRow,
        external_sample_id: str,
        cpg_sample_id: Optional[str],
        sequences: Dict[Any, Any],
    ):
        """
        ASYNC function that (maps) transforms one GroupedRow, and returns a Tuple of:
            (
                sample_to_upsert,
                sequences_to_upsert,
                analyses_to_add,
            )

        Then the calling function does the (reduce).
        """

        # now we have sample / sequencing meta across 4 different rows, so collapse them
        (
            collapsed_sequencing_meta,
            collapsed_sample_meta,
            collapsed_qc,
            collapsed_analyses,
        ) = await asyncio.gather(
            self.get_grouped_sequence_meta(external_sample_id, rows),
            self.get_sample_meta(SampleMetaGroup(sample_id=cpg_sample_id, rows=rows)),
            self.get_qc_meta(external_sample_id, rows[0]),
            self.get_analyses(external_sample_id, rows[0], cpg_id=cpg_sample_id),
        )

        sample_to_upsert = None
        sequences_to_upsert = []
        analyses_to_add = []

        # should we add or update sequencing
        if collapsed_sequencing_meta:
            for seq in collapsed_sequencing_meta:
                sequence_id = sequences.get(str(seq.sequence_type), None)
                if isinstance(sequence_id, list):
                    if len(sequence_id) > 1:
                        raise ValueError(
                            f'Unhandled case with more than one sequence ID for the type {seq.sequence_type}'
                        )
                    sequence_id = sequence_id[0]

                seq_args: Dict[str, Any] = {
                    'meta': seq.meta,
                    'type': seq.sequence_type,
                    'status': self.get_sequence_status(seq.rows),
                }

                if self.get_sequence_id(seq.rows):
                    seq_args['external_ids'] = self.get_sequence_id(seq.rows)

                if sequence_id:
                    seq_args['id'] = sequence_id

                sequences_to_upsert.append(SequenceUpsert(**seq_args))

        # Should we add or update sample
        sample_args: Dict[str, Any] = {
            'meta': collapsed_sample_meta.meta,
            'external_id': external_sample_id,
            'type': self.get_sample_type(rows),
            'sequences': sequences_to_upsert,
        }
        if cpg_sample_id:
            sample_args['id'] = cpg_sample_id

        sample_to_upsert = SampleBatchUpsert(**sample_args)

        if collapsed_analyses:
            analyses_to_add.extend(collapsed_analyses)

        if collapsed_qc:
            analyses_to_add.append(
                AnalysisModel(
                    sample_ids=['<none>'],
                    type=self.get_analysis_type(external_sample_id, rows),
                    status=self.get_analysis_status(external_sample_id, rows),
                    meta=collapsed_qc,
                )
            )

        return (
            sample_to_upsert,
            sequences_to_upsert,
            analyses_to_add,
        )

    async def process_participant_group(
        self,
        participant_name: str,
        sample_map,
        external_to_internal_sample_id_map: Dict[str, Any],
        internal_pid: Optional[int],
        sequence_map: Dict[str, Dict[str, int]] = None,
    ):
        """
        ASYNC function that (maps) transforms one GroupedRow, and returns a Tuple of:
            (
                participants_to_upsert,
                samples_to_upsert,
                sequences_to_upsert,
                analyses_to_add,
            )

        Then the calling function does the (reduce).
        """

        all_rows = [r for row in sample_map.values() for r in row]

        # Get all the samples and sequences to upsert first
        samples_to_upsert = []
        sequences_to_upsert = []
        analyses_to_add = []

        for sample_id, rows in sample_map.items():
            cpg_id = external_to_internal_sample_id_map.get(sample_id, None)
            sample, seqs, analyses = await self.process_sample_group(
                rows=rows,
                external_sample_id=sample_id,
                cpg_sample_id=cpg_id,
                sequences=sequence_map.get(cpg_id, {}),
            )
            samples_to_upsert.append(sample)
            sequences_to_upsert.extend(seqs)
            analyses_to_add.extend(analyses)

        # pull relevant participant fields
        reported_sex = self.get_reported_sex(all_rows)
        reported_gender = self.get_reported_gender(all_rows)
        karyotype = self.get_karyotype(all_rows)

        # now we have sample / sequencing meta across 4 different rows, so collapse them
        collapsed_participant_meta = await self.get_participant_meta(
            internal_pid, all_rows
        )

        args: Dict[str, Any] = {
            'external_id': participant_name,
            'meta': collapsed_participant_meta.meta,
            'samples': samples_to_upsert,
        }
        if internal_pid:
            args['id'] = internal_pid

        if reported_sex:
            args['reported_sex'] = reported_sex
        if reported_gender:
            args['reported_gender'] = reported_gender
        if karyotype:
            args['karyotype'] = karyotype

        participant_to_upsert = ParticipantUpsert(**args)

        return (
            participant_to_upsert,
            samples_to_upsert,
            sequences_to_upsert,
            analyses_to_add,
        )

    async def file_pointer_to_sample_map(
        self,
        file_pointer,
        delimiter: str,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        Parse manifest file into a list of dicts, indexed by sample name.
        """
        sample_map = defaultdict(list)
        reader = self._get_dict_reader(file_pointer, delimiter=delimiter)
        for row in reader:
            sid = self.get_sample_id(row)
            sample_map[sid].append(row)
        return sample_map

    def _get_dict_reader(self, file_pointer, delimiter: str):
        """
        Return a DictReader from file_pointer
        Override this method if you can't use the default implementation that simply
        calls csv.DictReader
        """
        reader = CustomDictReader(
            file_pointer,
            delimiter=delimiter,
            key_map=self.key_map,
            required_keys=self.required_keys,
            ignore_extra_keys=self.ignore_extra_keys,
        )
        return reader

    async def file_pointer_to_participant_map(
        self,
        file_pointer,
        delimiter: str,
    ) -> Dict[Any, Dict[str, List[Dict[str, Any]]]]:
        """
        Parse manifest file into a list of dicts, indexed by participant id.
        """

        participant_map: Dict[str, Dict[Any, List[Any]]] = defaultdict(
            lambda: defaultdict(list)
        )
        reader = self._get_dict_reader(file_pointer, delimiter=delimiter)
        for row in reader:
            participant_eid = self.get_participant_id(row)

            if not participant_eid:
                raise ValueError(f'Participant not found in row: {row}')

            # business rule to default sample ID to participant ID if not provided
            sid = self.get_sample_id(row) or participant_eid
            participant_map[participant_eid][sid].append(row)

        return participant_map

    async def validate_participant_map(
        self, participant_map: Dict[Any, Dict[str, List[Dict[str, Any]]]]
    ):
        """
        Validate sample rows:
        - throw an exception if an error occurs
        - log a warning for all other issues
        """
        if len(participant_map) == 0:
            raise ValueError('The manifest contains no records')

    async def validate_sample_map(self, sample_map: Dict[str, List[Dict[str, Any]]]):
        """
        Validate sample rows:
        - throw an exception if an error occurs
        - log a warning for all other issues
        """
        if len(sample_map) == 0:
            raise ValueError('The manifest contains no records')

    async def parse_manifest(  # pylint: disable=too-many-branches
        self, file_pointer, delimiter=',', confirm=False, dry_run=False
    ) -> Dict[str, Dict]:
        """
        Parse manifest from iterable (file pointer / String.IO)

        Returns a dict mapping external sample ID to CPG sample ID
        """
        if self.has_participants(file_pointer, delimiter):
            participant_map = await self.file_pointer_to_participant_map(
                file_pointer, delimiter
            )
            await self.validate_participant_map(participant_map)
            return await self.parse_manifest_by_participants(
                participant_map, confirm=confirm, dry_run=dry_run
            )

        sample_map = await self.file_pointer_to_sample_map(file_pointer, delimiter)
        await self.validate_sample_map(sample_map)

        return await self.parse_manifest_by_samples(
            sample_map, confirm=confirm, dry_run=dry_run
        )

    def upsert_summary(
        self,
        existing_summary: Dict[str, Dict[str, List[Any]]],
        participants_to_upsert: List[ParticipantUpsert],
        samples_to_upsert: List[SampleBatchUpsert],
        sequences_to_upsert: List[SequenceUpsert],
        analyses_to_add: Dict[str, List[AnalysisModel]],
    ) -> Dict[str, Dict[str, List[Any]]]:
        """Given lists of values to upsert return grouped summary of updates and inserts"""

        def upsert_type(item: Dict[str, Any]) -> str:
            return 'update' if hasattr(item, 'id') else 'insert'

        # Set initial dictionary value
        summary: Dict[str, Dict[str, List[Any]]] = (
            existing_summary.copy()
            if existing_summary
            else defaultdict(lambda: defaultdict(list))
        )

        for participant in participants_to_upsert:
            summary['participants'][upsert_type(participant)].append(participant)

        for sample in samples_to_upsert:
            summary['samples'][upsert_type(sample)].append(sample)

        for sequence in sequences_to_upsert:
            summary['sequences'][upsert_type(sequence)].append(sequence)

        for sid, analyses in analyses_to_add.items():
            summary['analyses'][sid].extend(analyses)

        return summary

    async def add_analyses(self, analyses_to_add, external_to_internal_id_map):
        """Given an analyses dictionary add analyses"""
        proj = self.project
        analysisapi = AnalysisApi()

        logger.info(
            f'{proj}: Adding analysis entries for {len(analyses_to_add)} samples'
        )
        unwrapped_analysis_to_add = [
            (sample_id, a)
            for (sample_id, analyses) in analyses_to_add.items()
            for a in analyses
        ]

        results = []
        for chunked_analysis in chunk(unwrapped_analysis_to_add):
            promises = []
            for external_id, analysis in chunked_analysis:
                analysis.sample_ids = [external_to_internal_id_map[external_id]]
                promises.append(
                    analysisapi.create_new_analysis_async(
                        project=proj, analysis_model=analysis
                    )
                )
            results.append(await asyncio.gather(*promises))

        return results

    async def parse_manifest_by_participants(
        self,
        participant_map: Dict[str, Any],
        confirm: bool = False,
        dry_run: bool = False,
    ) -> Dict[str, Any]:
        """Parses a manifest of data that is keyed on participant id and sample id"""

        proj = self.project

        # all dicts indexed by external_sample_id
        summary: Dict[str, Dict[str, List[Any]]] = None
        all_participants: List[ParticipantUpsert] = []
        analyses_to_add: Dict[str, List[AnalysisModel]] = defaultdict(list)

        # we'll batch process the samples as not to open too many threads
        for external_pids in chunk(list(participant_map.keys())):

            current_batch_promises = {}
            if self.verbose:
                logger.info(f'{proj}:Preparing {", ".join(external_pids)}')

            # Construct participant to upsert
            existing_participant_ids = (
                await self.papi.get_participant_id_map_by_external_ids_async(
                    self.project, external_pids, allow_missing=True
                )
            )

            sample_ids = list(
                set(k for s in participant_map.values() for k in s.keys())
            )

            external_sequence_ids = self.get_existing_external_sequence_ids(
                participant_map
            )

            external_to_internal_sample_id_map = (
                await self.sapi.get_sample_id_map_by_external_async(
                    self.project, sample_ids, allow_missing=True
                )
            )

            sequences = []
            existing_sequences: list[dict[str, Any]] = []
            if external_to_internal_sample_id_map:
                sequences = await self.seqapi.get_sequences_by_sample_ids_async(
                    list(external_to_internal_sample_id_map.values()),
                )

                # Accounts for multiple valid sequences per sample.
                existing_sequences = self.get_existing_sequences(
                    sequences, external_sequence_ids
                )

            sequence_map: Dict[str, Dict[str, int]] = defaultdict(dict)
            for seq in existing_sequences:
                sequence_map[seq['sample_id']][seq['type']] = seq['id']

            for external_pid in external_pids:
                sample_map = participant_map[external_pid]
                promise = self.process_participant_group(
                    participant_name=external_pid,
                    sample_map=sample_map,
                    internal_pid=existing_participant_ids.get(external_pid),
                    sequence_map=sequence_map,
                    external_to_internal_sample_id_map=external_to_internal_sample_id_map,
                )
                current_batch_promises[external_pid] = promise

            processed_ex_pids = list(current_batch_promises.keys())
            batch_promises = list(current_batch_promises.values())
            resolved_promises = await asyncio.gather(*batch_promises)

            for external_pid, resolved_promise in zip(
                processed_ex_pids, resolved_promises
            ):
                (
                    participant_to_upsert,
                    samples_to_upsert,
                    sequences_to_upsert,
                    analysis_to_add,
                ) = resolved_promise

                if analysis_to_add:
                    analyses_to_add[external_pid] = analysis_to_add

                if participant_to_upsert:
                    all_participants.append(participant_to_upsert)

                # Get summary information
                summary = self.upsert_summary(
                    summary,
                    [participant_to_upsert],
                    samples_to_upsert,
                    sequences_to_upsert,
                    {external_pid: analysis_to_add},
                )

        sequences_count: Dict[str, int] = defaultdict(int)
        for s in summary['sequences']['insert'] + summary['sequences']['update']:
            sequences_count[str(s['type'])] += 1

        str_seq_count = ', '.join(f'{k}={v}' for k, v in sequences_count.items())
        message = f"""\
    {proj}: Processing participants: {', '.join(participant_map.keys())}

    Sequence types: {str_seq_count}

    Adding {len(summary['participants']['insert'])} participants
    Adding {len(summary['samples']['insert'])} samples
    Adding {len(summary['sequences']['insert'])} sequences
    Adding {len(summary['analyses']['insert'])} analysis

    Updating {len(summary['participants']['update'])} participants
    Updating {len(summary['samples']['update'])} samples
    Updating {len(summary['sequences']['update'])} sequences
        """

        if dry_run:
            logger.info('Dry run, so returning without inserting / updating metadata')
            return summary

        if confirm:
            resp = str(input(message + '\n\nConfirm (y): '))
            if resp.lower() != 'y':
                raise SystemExit()
        else:
            logger.info(message)

        # Batch update
        result = self.papi.batch_upsert_participants(
            proj, ParticipantUpsertBody(participants=all_participants)
        )

        return result

    async def parse_manifest_by_samples(
        self, sample_map: Dict[str, Any], confirm: bool = False, dry_run: bool = False
    ) -> Dict[str, Any]:
        """Parses a manifest of data that is keyed on sample id"""
        proj = self.project

        # now we can start adding!!

        # Map external sids into cpg ids

        external_to_internal_sample_id_map = (
            await self.sapi.get_sample_id_map_by_external_async(
                self.project, list(sample_map.keys()), allow_missing=True
            )
        )

        sequences = []
        if external_to_internal_sample_id_map:
            sequences = await self.seqapi.get_sequences_by_sample_ids_async(
                list(external_to_internal_sample_id_map.values()),
            )

        sequence_map: Dict[str, Dict[str, int]] = defaultdict(dict)
        for seq in sequences:
            sequence_map[seq['sample_id']][seq['type']] = seq['id']

        # all dicts indexed by external_sample_id
        summary: Dict[str, Dict[str, List[Any]]] = None
        all_samples: List[SampleBatchUpsert] = []
        analyses_to_add: Dict[str, List[AnalysisModel]] = defaultdict(list)

        # we'll batch process the samples as not to open too many threads
        for external_sids in chunk(list(sample_map.keys())):

            current_batch_promises = {}
            if self.verbose:
                logger.info(f'{proj}:Preparing {", ".join(external_sids)}')

            for external_sid in external_sids:
                rows: GroupedRow = sample_map[external_sid]
                cpg_id = external_to_internal_sample_id_map.get(external_sid)
                promise = self.process_sample_group(
                    rows=rows,
                    external_sample_id=external_sid,
                    cpg_sample_id=cpg_id,
                    sequences=sequence_map[cpg_id],
                )
                current_batch_promises[external_sid] = promise

            processed_ex_sids = list(current_batch_promises.keys())
            batch_promises = list(current_batch_promises.values())
            resolved_promises = await asyncio.gather(*batch_promises)

            for external_sid, resolved_promise in zip(
                processed_ex_sids, resolved_promises
            ):
                (
                    sample_to_upsert,
                    sequences_to_upsert,
                    analysis_to_add,
                ) = resolved_promise

                # Extract Upsert items and add to an array
                if analysis_to_add:
                    analyses_to_add[external_sid] = analysis_to_add

                if sample_to_upsert:
                    all_samples.append(sample_to_upsert)

                # Get summary information
                summary = self.upsert_summary(
                    summary,
                    [],
                    [sample_to_upsert],
                    sequences_to_upsert,
                    {external_sid: analysis_to_add},
                )

        message = f"""\
            {proj}: Processing samples: {', '.join(sample_map.keys())}

            Adding {len(summary['samples']['insert'])} samples
            Adding {len(summary['sequences']['insert'])} sequences
            Adding {len(summary['analyses']['insert'])} analyses

            Updating {len(summary['samples']['update'])} samples
            Updating {len(summary['sequences']['update'])} sequences
        """

        if dry_run:
            logger.info('Dry run, so returning without inserting / updating metadata')
            return summary

        if confirm:
            resp = str(input(message + '\n\nConfirm (y): '))
            if resp.lower() != 'y':
                raise SystemExit()
        else:
            logger.info(message)

        # Batch update
        result = await self.sapi.batch_upsert_samples_async(
            proj, SampleBatchUpsertBody(samples=all_samples)
        )

        # Add analyses
        # Map external sids into cpg ids
        existing_external_id_to_cpgid = (
            await self.sapi.get_sample_id_map_by_external_async(
                proj, list(sample_map.keys()), allow_missing=True
            )
        )
        _ = await self.add_analyses(analyses_to_add, existing_external_id_to_cpgid)

        return result

    async def parse_files(
        self, sample_id: str, reads: List[str], checksums: List[str] = None
    ) -> Dict[SUPPORTED_FILE_TYPE, Dict[str, List]]:
        """
        Returns a tuple of:
        1. single / list-of CWL file object(s), based on the extensions of the reads
        2. parsed type (fastq, cram, bam)
        """

        if not isinstance(reads, list):
            reads = [reads]

        if not checksums:
            checksums = [None] * len(reads)

        if len(checksums) != len(reads):
            raise ValueError(
                'Expected length of reads to match length of provided checksums'
            )

        read_to_checksum = dict(zip(reads, checksums))

        file_by_type: Dict[SUPPORTED_FILE_TYPE, Dict[str, List]] = defaultdict(
            lambda: defaultdict(list)
        )

        fastqs = [
            r for r in reads if any(r.lower().endswith(ext) for ext in FASTQ_EXTENSIONS)
        ]
        if fastqs:
            structured_fastqs = self.parse_fastqs_structure(fastqs)
            fastq_files: List[Sequence[Union[Coroutine, BaseException]]] = []  # type: ignore
            for fastq_group in structured_fastqs:
                create_file_futures: List[Coroutine] = [
                    self.create_file_object(f, checksum=read_to_checksum.get(f))
                    for f in fastq_group
                ]
                fastq_files.append(asyncio.gather(*create_file_futures))  # type: ignore

            grouped_fastqs = list(await asyncio.gather(*fastq_files))  # type: ignore
            file_by_type['reads']['fastq'].extend(grouped_fastqs)

        crams = [
            r for r in reads if any(r.lower().endswith(ext) for ext in CRAM_EXTENSIONS)
        ]
        file_promises: List[Coroutine]

        if crams:
            file_promises = []
            sec_format = ['.crai', '^.crai']
            for r in crams:
                secondaries = (
                    await self.create_secondary_file_objects_by_potential_pattern(
                        r, sec_format
                    )
                )
                file_promises.append(
                    self.create_file_object(r, secondary_files=secondaries)
                )
            file_by_type['reads']['cram'] = await asyncio.gather(*file_promises)  # type: ignore

        bams = [
            r for r in reads if any(r.lower().endswith(ext) for ext in BAM_EXTENSIONS)
        ]
        if bams:
            file_promises = []
            sec_format = ['.bai', '^.bai']
            for r in bams:
                secondaries = (
                    await self.create_secondary_file_objects_by_potential_pattern(
                        r, sec_format
                    )
                )
                file_promises.append(
                    self.create_file_object(r, secondary_files=secondaries)
                )

            file_by_type['reads']['bam'] = await asyncio.gather(*file_promises)  # type: ignore

        gvcfs = [
            r for r in reads if any(r.lower().endswith(ext) for ext in GVCF_EXTENSIONS)
        ]
        vcfs = [
            r
            for r in reads
            if any(r.lower().endswith(ext) for ext in VCF_EXTENSIONS) and r not in gvcfs
        ]

        if gvcfs:
            file_promises = []
            sec_format = ['.tbi']
            for r in gvcfs:
                secondaries = (
                    await self.create_secondary_file_objects_by_potential_pattern(
                        r, sec_format
                    )
                )
                file_promises.append(
                    self.create_file_object(r, secondary_files=secondaries)
                )

            file_by_type['variants']['gvcf'] = await asyncio.gather(*file_promises)  # type: ignore

        if vcfs:
            file_promises = []
            for r in vcfs:
                file_promises.append(self.create_file_object(r))

            file_by_type['variants']['vcf'] = await asyncio.gather(*file_promises)  # type: ignore

        unhandled_files = [
            r
            for r in reads
            if not any(r.lower().endswith(ext) for ext in ALL_EXTENSIONS)
        ]
        if unhandled_files:
            joined_reads = ''.join(
                f'\n\t{i}: {r}' for i, r in enumerate(unhandled_files)
            )
            logger.warning(
                f'There were files with extensions that were skipped ({sample_id}): {joined_reads}'
            )

        return file_by_type

    @staticmethod
    def parse_fastqs_structure(fastqs) -> List[List[str]]:
        """
        Takes a list of fastqs, and a set of nested lists of each R1 + R2 read.

        >>> GenericParser.parse_fastqs_structure(['/seqr_transfers/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L004_R1.fastq.gz', '/seqr_transfers/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L004_R2.fastq.gz'])
        [['/seqr_transfers/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L004_R1.fastq.gz', '/seqr_transfers/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L004_R2.fastq.gz']]

        >>> GenericParser.parse_fastqs_structure(['20210727_PROJECT1_L002_R2.fastq.gz', '20210727_PROJECT1_L002_R1.fastq.gz', '20210727_PROJECT1_L001_R2.fastq.gz', '20210727_PROJECT1_L001_R1.fastq.gz'])
        [['20210727_PROJECT1_L001_R1.fastq.gz', '20210727_PROJECT1_L001_R2.fastq.gz'], ['20210727_PROJECT1_L002_R1.fastq.gz', '20210727_PROJECT1_L002_R2.fastq.gz']]

        >>> GenericParser.parse_fastqs_structure(['/directory1/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R1.fastq.gz', '/directory2/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R2.fastq.gz'])
        [['/directory1/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R1.fastq.gz', '/directory2/Z01_1234_HNXXXXX_TCATCCTT-AGCGAGCT_L001_R2.fastq.gz']]

        >>> GenericParser.parse_fastqs_structure(['Sample_1_L01_1.fastq.gz', 'Sample_1_L01_2.fastq.gz', 'Sample_1_L02_R1.fastq.gz', 'Sample_1_L02_R2.fastq.gz'])
        [['Sample_1_L01_1.fastq.gz', 'Sample_1_L01_2.fastq.gz'], ['Sample_1_L02_R1.fastq.gz', 'Sample_1_L02_R2.fastq.gz']]

        >>> GenericParser.parse_fastqs_structure(['File_1.fastq', 'File_2.fastq'])
        [['File_1.fastq', 'File_2.fastq']]

        >>> GenericParser.parse_fastqs_structure(['File_1.fq', 'File_2.fq'])
        [['File_1.fq', 'File_2.fq']]

        >>> GenericParser.parse_fastqs_structure(['File_1.fq.gz', 'File_2.fq.gz'])
        [['File_1.fq.gz', 'File_2.fq.gz']]
        """
        # find last instance of R\d, and then group by prefix on that
        sorted_fastqs = sorted(fastqs)

        r_matches: Dict[str, Tuple[str, Optional[Match[str]]]] = {
            r: (os.path.basename(r), rmatch.search(os.path.basename(r)))
            for r in sorted_fastqs
        }
        no_r_match = [r for r, (_, matched) in r_matches.items() if matched is None]
        if no_r_match:
            no_r_match_str = ', '.join(no_r_match)
            raise ValueError(
                f"Couldn't detect the format of FASTQs (expected match for regex '{rmatch.pattern}'): {no_r_match_str}"
            )

        values = []
        for _, grouped in group_by(
            sorted_fastqs, lambda r: r_matches[r][0][: r_matches[r][1].start()]  # type: ignore
        ).items():
            values.append(sorted(grouped))
        invalid_fastq_groups = [grp for grp in values if len(grp) != 2]
        if invalid_fastq_groups:
            raise ValueError(f'Invalid fastq group {invalid_fastq_groups}')

        return sorted(values, key=lambda el: el[0])

    async def create_file_object(
        self,
        filename: str,
        secondary_files: List[SingleRow] = None,
        checksum: Optional[str] = None,
    ) -> SingleRow:
        """Takes filename, returns formed CWL dictionary"""
        _checksum = checksum
        file_size = None

        if not self.skip_checking_gcs_objects:
            if not _checksum:
                md5_filename = self.file_path(filename + '.md5')
                if await self.file_exists(md5_filename):
                    contents = await self.file_contents(md5_filename)
                    if contents:
                        _checksum = f'md5:{contents.strip()}'

            file_size = await self.file_size(filename)

        d = {
            'location': self.file_path(filename),
            'basename': os.path.basename(filename),
            'class': 'File',
            'checksum': _checksum,
            'size': file_size,
        }

        if secondary_files:
            d['secondaryFiles'] = secondary_files

        return d

    async def create_secondary_file_objects_by_potential_pattern(
        self, filename, potential_secondary_patterns: List[str]
    ) -> List[SingleRow]:
        """
        Take a base filename and potential secondary patterns:
        - Try each secondary pattern, see if it works
        - If it works, create a CWL file object
        - return a list of those secondary file objects that exist
        """
        secondaries = []
        for sec in potential_secondary_patterns:
            sec_file = _apply_secondary_file_format_to_filename(filename, sec)
            if self.skip_checking_gcs_objects or await self.file_exists(sec_file):
                secondaries.append(self.create_file_object(sec_file))

        return await asyncio.gather(*secondaries)

    @staticmethod
    def guess_delimiter_from_filename(filename: str):
        """
        Guess delimiter from filename
        """
        extension_to_delimiter = {'.csv': ',', '.tsv': '\t'}
        relevant_delimiter = next(
            (
                delimiter
                for ext, delimiter in extension_to_delimiter.items()
                if filename.endswith(ext)
            ),
            None,
        )
        if relevant_delimiter:
            return relevant_delimiter

        # pylint: disable=no-member
        with AnyPath(filename).open('r') as f:
            first_line = f.readline()
            delimiter = csv.Sniffer().sniff(first_line).delimiter
            if delimiter:
                logger.info(
                    f'Guessing delimiter based on first line, got "{delimiter}"'
                )
                return delimiter

        raise ValueError(f'Unrecognised extension on file: {filename}')


def _apply_secondary_file_format_to_filename(
    filepath: Optional[str], secondary_file: str
):
    """
    You can trust this function to do what you want
    :param filepath: Filename to base
    :param secondary_file: CWL secondary format (Remove 1 extension for each leading ^).
    """
    if not filepath:
        return None

    fixed_sec = secondary_file.lstrip('^')
    leading = len(secondary_file) - len(fixed_sec)
    if leading <= 0:
        return filepath + fixed_sec

    basepath = ''
    filename = filepath
    if '/' in filename:
        idx = len(filepath) - filepath[::-1].index('/')
        basepath = filepath[:idx]
        filename = filepath[idx:]

    split = filename.split('.')

    newfname = filename + fixed_sec
    if len(split) > 1:
        newfname = '.'.join(split[: -min(leading, len(split) - 1)]) + fixed_sec
    return basepath + newfname



# pylint: disable=invalid-overridden-method

import asyncio
import logging
import os
import socket
import subprocess
import unittest
from functools import wraps

from typing import Dict

from pymysql import IntegrityError
from testcontainers.mysql import MySqlContainer
import nest_asyncio

from db.python.connect import (
    ConnectionStringDatabaseConfiguration,
    Connection,
    SMConnections,
    TABLES_ORDERED_BY_FK_DEPS,
)
from db.python.tables.project import ProjectPermissionsTable, set_full_access

# use this to determine where the db directory is relatively,
# as pycharm runs in "test/" folder, and GH runs them in git root
am_i_in_test_environment = os.getcwd().endswith('test')

# This monkey patches the asyncio event loop and allows it to be re-entrant
# (you can call run_until_complete while run_until_complete is already on the stack)
# Source: https://stackoverflow.com/a/56434301
nest_asyncio.apply()


for lname in (
    'asyncio',
    'urllib3',
    'docker',
    'databases',
    'testcontainers.core.container',
    'testcontainers.core.waiting_utils',
):
    logging.getLogger(lname).setLevel(logging.WARNING)


def find_free_port():
    """Find free port to run tests on"""
    s = socket.socket()
    s.bind(('', 0))  # Bind to a free port provided by the host.
    return s.getsockname()[1]  # Return the port number assigned.


loop = asyncio.new_event_loop()


def run_as_sync(f):
    """
    Decorate your async function to run is synchronously
    This runs on the testing event loop
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


class DbTest(unittest.TestCase):
    """Base class for database integration tests"""

    # store connections here, so they can be created PER-CLASS
    # and don't get recreated per test.
    dbs: Dict[str, MySqlContainer] = {}
    connections: Dict[str, Connection] = {}

    @classmethod
    def setUpClass(cls) -> None:
        @run_as_sync
        async def setup():
            """
            This starts a mariadb container, applies liquibase schema and inserts a project
            MAYBE, the best way in the future would be to only ever create ONE connection
            between all Test classes, and create a new database per test, new set of a connections,
            but that certaintly has a host of it's own problems

            Then you can destroy the database within tearDownClass as all tests have been completed.
            """
            logger = logging.getLogger()
            try:
                os.environ['SM_ALLOWALLACCESS'] = '1'
                set_full_access(True)
                db = MySqlContainer('mariadb:10.8.3')
                port_to_expose = find_free_port()
                # override the default port to map the container to
                db.with_bind_ports(db.port_to_expose, port_to_expose)
                logger.disabled = True
                db.start()
                logger.disabled = False
                cls.dbs[cls.__name__] = db

                db_prefix = 'db'
                if am_i_in_test_environment:
                    db_prefix = '../db'

                con_string = db.get_connection_url()
                con_string = 'mysql://' + con_string.split('://', maxsplit=1)[1]
                lcon_string = f'jdbc:mariadb://{db.get_container_host_ip()}:{port_to_expose}/{db.MYSQL_DATABASE}'
                # apply the liquibase schema
                command = [
                    'liquibase',
                    *('--changeLogFile', db_prefix + '/project.xml'),
                    *('--defaultsFile', db_prefix + '/liquibase.properties'),
                    *('--url', lcon_string),
                    *('--driver', 'org.mariadb.jdbc.Driver'),
                    *('--classpath', db_prefix + '/mariadb-java-client-3.0.3.jar'),
                    *('--username', db.MYSQL_USER),
                    *('--password', db.MYSQL_PASSWORD),
                    'update',
                ]
                subprocess.check_output(command, stderr=subprocess.STDOUT)

                sm_db = SMConnections.make_connection(
                    ConnectionStringDatabaseConfiguration(con_string)
                )
                await sm_db.connect()
                connection = Connection(
                    connection=sm_db,
                    project=1,
                    author='testuser',
                )
                cls.connections[cls.__name__] = connection

                ppt = ProjectPermissionsTable(
                    connection.connection, allow_full_access=True
                )
                await ppt.create_project(
                    project_name='test',
                    dataset_name='test',
                    create_test_project=False,
                    gcp_project_id='None',
                    author='testuser',
                    read_secret_name='None',
                    write_secret_name='None',
                    check_permissions=False,
                )

            except subprocess.CalledProcessError as e:
                logging.exception(e)
                if e.stderr:
                    logging.error(e.stderr.decode().replace('\\n', '\n'))
                if e.stdout:
                    logging.error(e.stdout.decode().replace('\\n', '\n'))
                cls.tearDownClass()
                raise e
            except Exception as e:
                logging.error(f'FAILED WITH {e}')
                logging.exception(e)
                cls.tearDownClass()

                raise

        return setup()

    @classmethod
    def tearDownClass(cls) -> None:
        db = cls.dbs.get(cls.__name__)
        if db:
            db.exec(f'DROP DATABASE {db.MYSQL_DATABASE};')
            db.stop()

    def setUp(self) -> None:
        self.project_id = 1
        self.connection = self.connections[self.__class__.__name__]


class DbIsolatedTest(DbTest):
    """
    Database integration tests that performs clean-up at the start of each test
    """

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        ignore = {'DATABASECHANGELOG', 'DATABASECHANGELOGLOCK', 'project'}
        for table in TABLES_ORDERED_BY_FK_DEPS:
            if table in ignore:
                continue
            try:
                await self.connection.connection.execute(f'DELETE FROM {table}')
                await self.connection.connection.execute(f'DELETE HISTORY FROM {table}')
            except IntegrityError as e:
                raise IntegrityError(f'Could not delete {table}') from e



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.search import SearchLayer
from db.python.layers.family import FamilyLayer
from db.python.tables.family_participant import FamilyParticipantTable

from models.enums import SampleType, SearchResponseType
from models.models.sample import sample_id_format


class TestSample(DbIsolatedTest):
    """Test sample class"""

    # tests run in 'sorted by ascii' order
    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.schlay = SearchLayer(self.connection)
        self.slayer = SampleLayer(self.connection)
        self.player = ParticipantLayer(self.connection)
        self.flayer = FamilyLayer(self.connection)

    @run_as_sync
    async def test_search_non_existent_sample_by_internal_id(self):
        """
        Search by CPG sample ID that doesn't exist
        """
        cpg_id = sample_id_format(9_999_999_999)
        results = await self.schlay.search(query=cpg_id, project_ids=[self.project_id])
        self.assertEqual(0, len(results))

    @run_as_sync
    async def test_search_unavailable_sample_by_internal_id(self):
        """
        Search by CPG sample ID that you do not have access to
        Mock this in testing by limiting scope to non-existent project IDs
        """
        sample_id = await self.slayer.insert_sample('EX001', SampleType.BLOOD)
        cpg_id = sample_id_format(sample_id)

        results = await self.schlay.search(
            query=cpg_id, project_ids=[self.project_id + 1]
        )
        self.assertEqual(1, len(results))

    @run_as_sync
    async def test_search_isolated_sample_by_id(self):
        """
        Search by valid CPG sample ID (special case)
        """
        sample_id = await self.slayer.insert_sample('EX001', SampleType.BLOOD)
        cpg_id = sample_id_format(sample_id)
        results = await self.schlay.search(query=cpg_id, project_ids=[self.project_id])

        self.assertEqual(1, len(results))
        self.assertEqual(cpg_id, results[0].title)
        self.assertEqual(cpg_id, results[0].data.id)
        self.assertListEqual(['EX001'], results[0].data.sample_external_ids)

    @run_as_sync
    async def test_search_isolated_sample_by_external_id(self):
        """
        Search by External sample ID with no participant / family,
        should only return one result
        """
        sample_id = await self.slayer.insert_sample('EX001', SampleType.BLOOD)
        results = await self.schlay.search(query='EX001', project_ids=[self.project_id])

        cpg_id = sample_id_format(sample_id)

        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual(cpg_id, result.title)
        self.assertEqual(cpg_id, result.data.id)
        self.assertListEqual(['EX001'], result.data.sample_external_ids)
        self.assertListEqual([], result.data.participant_external_ids)
        self.assertListEqual([], result.data.family_external_ids)

    @run_as_sync
    async def test_search_participant_isolated(self):
        """
        Search participant w/ no family by External ID
        should only return one result
        """
        p_id = await self.player.create_participant(external_id='PART01')
        results = await self.schlay.search(
            query='PART01', project_ids=[self.project_id]
        )
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual(p_id, result.data.id)
        self.assertEqual('PART01', result.title)
        self.assertListEqual(['PART01'], result.data.participant_external_ids)
        self.assertListEqual([], result.data.family_external_ids)
        self.assertRaises(AttributeError, lambda: result.data.sample_external_ids)

    @run_as_sync
    async def test_search_family(self):
        """
        Search family by External ID
        should only return one result
        """
        f_id = await self.flayer.create_family(external_id='FAMXX01')
        results = await self.schlay.search(
            query='FAMXX01', project_ids=[self.project_id]
        )
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual(f_id, result.data.id)
        self.assertEqual('FAMXX01', result.title)
        self.assertListEqual(['FAMXX01'], result.data.family_external_ids)
        self.assertRaises(AttributeError, lambda: result.data.participant_external_ids)
        self.assertRaises(AttributeError, lambda: result.data.sample_external_ids)

    @run_as_sync
    async def test_search_mixed(self):
        """Create a number of resources, and search for all of them"""
        fptable = FamilyParticipantTable(self.connection)

        p_id = await self.player.create_participant(external_id='X:PART01')
        f_id = await self.flayer.create_family(external_id='X:FAM01')
        await fptable.create_rows(
            [
                {
                    'family_id': f_id,
                    'participant_id': p_id,
                    'paternal_participant_id': None,
                    'maternal_participant_id': None,
                    'affected': 0,
                    'notes': None,
                }
            ]
        )

        s_id = await self.slayer.insert_sample(
            'X:SAM001', SampleType.BLOOD, participant_id=p_id
        )

        all_results = await self.schlay.search(
            query='X:', project_ids=[self.connection.project]
        )
        self.assertEqual(3, len(all_results))

        family_result = next(
            r for r in all_results if r.type == SearchResponseType.FAMILY
        )
        participant_result = next(
            r for r in all_results if r.type == SearchResponseType.PARTICIPANT
        )
        sample_result = next(
            r for r in all_results if r.type == SearchResponseType.SAMPLE
        )

        # linked family matches
        self.assertEqual('X:FAM01', family_result.title)

        # linked participant matches
        self.assertEqual('X:PART01', participant_result.title)
        self.assertListEqual(['X:FAM01'], participant_result.data.family_external_ids)

        # linked sample matches
        cpg_id = sample_id_format(s_id)
        self.assertEqual(cpg_id, sample_result.data.id)
        self.assertListEqual(['X:SAM001'], sample_result.data.sample_external_ids)
        self.assertListEqual(['X:FAM01'], participant_result.data.family_external_ids)
        self.assertListEqual(
            ['X:PART01'], participant_result.data.participant_external_ids
        )



# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from sample_metadata.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from sample_metadata.model.analysis_model import AnalysisModel
from sample_metadata.model.analysis_query_model import AnalysisQueryModel
from sample_metadata.model.analysis_status import AnalysisStatus
from sample_metadata.model.analysis_type import AnalysisType
from sample_metadata.model.analysis_update_model import AnalysisUpdateModel
from sample_metadata.model.body_get_latest_complete_analysis_for_type_post import BodyGetLatestCompleteAnalysisForTypePost
from sample_metadata.model.body_get_participants import BodyGetParticipants
from sample_metadata.model.body_get_samples import BodyGetSamples
from sample_metadata.model.body_get_sequences_by_criteria import BodyGetSequencesByCriteria
from sample_metadata.model.export_type import ExportType
from sample_metadata.model.extra_participant_importer_handler import ExtraParticipantImporterHandler
from sample_metadata.model.family_search_response_data import FamilySearchResponseData
from sample_metadata.model.family_update_model import FamilyUpdateModel
from sample_metadata.model.http_validation_error import HTTPValidationError
from sample_metadata.model.nested_family import NestedFamily
from sample_metadata.model.nested_participant import NestedParticipant
from sample_metadata.model.nested_sample import NestedSample
from sample_metadata.model.nested_sequence import NestedSequence
from sample_metadata.model.new_sample import NewSample
from sample_metadata.model.new_sequence import NewSequence
from sample_metadata.model.paging_links import PagingLinks
from sample_metadata.model.participant_search_response_data import ParticipantSearchResponseData
from sample_metadata.model.participant_update_model import ParticipantUpdateModel
from sample_metadata.model.participant_upsert import ParticipantUpsert
from sample_metadata.model.participant_upsert_body import ParticipantUpsertBody
from sample_metadata.model.project import Project
from sample_metadata.model.project_summary_response import ProjectSummaryResponse
from sample_metadata.model.sample_batch_upsert import SampleBatchUpsert
from sample_metadata.model.sample_batch_upsert_body import SampleBatchUpsertBody
from sample_metadata.model.sample_search_response_data import SampleSearchResponseData
from sample_metadata.model.sample_type import SampleType
from sample_metadata.model.sample_update_model import SampleUpdateModel
from sample_metadata.model.search_response import SearchResponse
from sample_metadata.model.search_response_model import SearchResponseModel
from sample_metadata.model.search_response_type import SearchResponseType
from sample_metadata.model.sequence_status import SequenceStatus
from sample_metadata.model.sequence_type import SequenceType
from sample_metadata.model.sequence_update_model import SequenceUpdateModel
from sample_metadata.model.sequence_upsert import SequenceUpsert
from sample_metadata.model.validation_error import ValidationError
from sample_metadata.model.web_project import WebProject



from datetime import date

from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.web import (
    WebLayer,
    ProjectSummary,
    NestedParticipant,
    NestedSample,
    NestedSequence,
    WebProject,
)
from db.python.layers.participant import (
    ParticipantLayer,
    ParticipantUpsert,
    ParticipantUpsertBody,
)
from db.python.layers.sample import SampleBatchUpsert, SampleLayer
from db.python.layers.sequence import SampleSequenceLayer, SequenceUpsert

from models.enums.sample import SampleType
from models.enums.sequencing import SequenceStatus, SequenceType


def data_to_class(data: dict | list) -> dict | list:
    """Convert the data into it's class using the _class field"""
    if isinstance(data, list):
        return [data_to_class(x) for x in data]

    if not isinstance(data, dict):
        return data

    cls = data.pop('_class', None)
    mapped_data = {k: data_to_class(v) for k, v in data.items()}

    if isinstance(cls, type):
        return cls(**mapped_data)

    return mapped_data


def merge(d1: dict, d2: dict):
    """Merges two dictionaries"""
    return dict(d1, **d2)


class TestWeb(DbIsolatedTest):
    """Test web class containing web endpoints"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()
        self.webl = WebLayer(self.connection)
        self.partl = ParticipantLayer(self.connection)
        self.sampl = SampleLayer(self.connection)
        self.seql = SampleSequenceLayer(self.connection)

    @run_as_sync
    async def test_project_summary(self):
        """Test getting the summary for a project"""
        result = await self.webl.get_project_summary(token=None)

        # Expect an empty project
        expected = ProjectSummary(
            project=WebProject(
                **{'id': 1, 'name': 'test', 'meta': {}, 'dataset': 'test'}
            ),
            total_samples=0,
            total_participants=0,
            total_sequences=0,
            batch_sequence_stats={},
            cram_seqr_stats={},
            participants=[],
            participant_keys=[],
            sample_keys=[],
            sequence_keys=[],
            seqr_links={},
        )

        self.assertEqual(expected, result)

        # Now add a participant with a sample and sequence
        data = [
            {
                '_class': ParticipantUpsert,
                'external_id': 'Demeter',
                'meta': {},
                'samples': [
                    {
                        '_class': SampleBatchUpsert,
                        'external_id': 'sample_id001',
                        'meta': {},
                        'type': SampleType.BLOOD,
                        'sequences': [
                            {
                                '_class': SequenceUpsert,
                                'type': SequenceType.GENOME,
                                'status': SequenceStatus.UPLOADED,
                                'meta': {
                                    'reads': [
                                        [
                                            {
                                                'basename': 'sample_id001.filename-R1.fastq.gz',
                                                'checksum': None,
                                                'class': 'File',
                                                'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                                'size': 111,
                                            },
                                            {
                                                'basename': 'sample_id001.filename-R2.fastq.gz',
                                                'checksum': None,
                                                'class': 'File',
                                                'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                                'size': 111,
                                            },
                                        ]
                                    ],
                                    'reads_type': 'fastq',
                                    'batch': 'M001',
                                },
                            },
                        ],
                    }
                ],
            },
        ]

        body = ParticipantUpsertBody(participants=data_to_class(data))
        await self.partl.batch_upsert_participants(participants=body)

        # Switch to response classes
        expected_data: list = [
            merge(
                participant,
                {
                    '_class': NestedParticipant,
                    'id': 1,
                    'families': [],
                    'samples': [
                        merge(
                            sample,
                            {
                                '_class': NestedSample,
                                'id': 1,
                                'created_date': str(date.today()),
                                'sequences': [
                                    merge(
                                        sequence,
                                        {
                                            '_class': NestedSequence,
                                            'id': 1,
                                        },
                                    )
                                    for sequence in sample.get('sequences')
                                ],
                            },
                        )
                        for sample in participant.get('samples')
                    ],
                },
            )
            for participant in data
        ]

        result = await self.webl.get_project_summary(token=None)

        expected = ProjectSummary(
            project=WebProject(
                **{'id': 1, 'name': 'test', 'meta': {}, 'dataset': 'test'}
            ),
            total_samples=1,
            total_participants=1,
            total_sequences=1,
            cram_seqr_stats={
                'genome': {
                    'Sequences': '1',
                    'Crams': '0',
                    'Seqr': '0',
                }
            },
            batch_sequence_stats={'M001': {'genome': '1'}},
            participants=data_to_class(expected_data),
            participant_keys=[('external_id', 'Participant ID')],
            sample_keys=[
                ('id', 'Sample ID'),
                ('external_id', 'External Sample ID'),
                ('created_date', 'Created date'),
            ],
            sequence_keys=sorted(
                [
                    ('type', 'type'),
                    ('meta.reads_type', 'reads_type'),
                    ('meta.batch', 'batch'),
                ]
            ),
            seqr_links={},
        )

        self.assertEqual(expected, result)



# flake8: noqa

"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 5.3.0
    Generated by: https://openapi-generator.tech
"""


__version__ = "1.0.0"

# import ApiClient
from sample_metadata.api_client import ApiClient

# import Configuration
from sample_metadata.configuration import Configuration

# import exceptions
from sample_metadata.exceptions import OpenApiException
from sample_metadata.exceptions import ApiAttributeError
from sample_metadata.exceptions import ApiTypeError
from sample_metadata.exceptions import ApiValueError
from sample_metadata.exceptions import ApiKeyError
from sample_metadata.exceptions import ApiException



from typing import List

from sample_metadata.apis import SampleApi, AnalysisApi
from sample_metadata.models import (
    NewSample,
    AnalysisModel,
    AnalysisUpdateModel,
    SampleType,
    AnalysisType,
    AnalysisStatus,
)


INPUT_PROJECT = 'test_input_project'
OUTPUT_PROJECT = 'test_output_project'


aapi = AnalysisApi()
sapi = SampleApi()


def _add_samples(test_run_id: str, project: str):
    """
    Add 3 samples: one with fastq input, one with CRAM input, one with GVCF input.
    :param test_run_id: to suffix sample names for uniqueness
    """
    s1 = NewSample(
        external_id=f'NA12878-from-fq-{test_run_id}',
        type=SampleType('blood'),
        meta={
            'reads': [
                [
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L001_R1.fq',
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L002_R1.fq',
                ],
                [
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L001_R2.fq',
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L002_R2.fq',
                ],
            ]
        },
    )
    s2 = NewSample(
        external_id=f'NA12878-from-cram-{test_run_id}',
        type=SampleType('blood'),
        meta={'reads': 'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878.cram'},
    )
    s3 = NewSample(
        external_id=f'NA12878-from-gvcf-{test_run_id}',
        type=SampleType('blood'),
        meta={'reads': 'gs://cpg-seqr-test/batches/NA12878-trio/NA12878.g.vcf.gz'},
    )
    sample_ids = [sapi.create_new_sample(project, s) for s in (s1, s2, s3)]
    print(f'Added samples {", ".join(sample_ids)}')
    return sample_ids


def _submit_analyses(samples: List, output_project: str, a_type: str):
    """
    Add or update analyses. Iterate over completed analyses,
    and submit next-step analyses
    """

    if a_type in ['gvcf', 'cram']:
        for s in samples:
            if a_type == 'gvcf':
                cram_analysis = aapi.get_latest_analysis_for_samples_and_type(
                    project=output_project,
                    analysis_type=AnalysisType('cram'),
                    request_body=[s['id']],
                )
                assert cram_analysis, s

            # completed_analysis = latest_by_type_and_sids.get(('cram', (s['id'],))),
            am = AnalysisModel(
                type=AnalysisType(a_type),
                output=f'result.{a_type}',
                status=AnalysisStatus('queued'),
                sample_ids=[s['id']],
            )
            aapi.create_new_analysis(project=output_project, analysis_model=am)

    elif a_type == 'joint-calling':
        for s in samples:
            gvcf_analysis = aapi.get_latest_analysis_for_samples_and_type(
                project=output_project,
                analysis_type=AnalysisType('gvcf'),
                request_body=[s['id']],
            )
            assert gvcf_analysis, s

        am = AnalysisModel(
            sample_ids=[s['id'] for s in samples],
            type=AnalysisType('joint-calling'),
            output='joint-called.vcf',
            status=AnalysisStatus('queued'),
        )
        aapi.create_new_analysis(project=output_project, analysis_model=am)


def _update_analyses(proj: str, new_status: str):
    """
    Update existing analyses status
    """
    analyses = aapi.get_incomplete_analyses(project=proj)
    print(f'Current analyses:')
    for a in analyses:
        print(a)

    if analyses:
        for a in analyses:
            print(f'Setting analysis {a} to {new_status}')
            aum = AnalysisUpdateModel(status=AnalysisStatus(new_status))
            aapi.update_analysis_status(
                analysis_id=a['id'],
                analysis_update_model=aum,
            )


def test_simulate_joint_calling_pipeline(
    input_project: str,
    output_project: str,
):
    """
    Simulates events of the joint-calling workflow
    """
    samples = sapi.get_samples(
        body_get_samples_by_criteria_api_v1_sample_post={
            'project_ids': [input_project],
            'active': True,
        }
    )

    print('Add/update analyses, reads -> cram')
    _submit_analyses(samples, output_project, a_type='cram')
    print()
    _update_analyses(output_project, new_status='in-progress')
    print()
    _update_analyses(output_project, new_status='completed')
    print()

    print('Add/update analyses, cram -> gvcf')
    _submit_analyses(samples, output_project, a_type='gvcf')
    print()
    _update_analyses(output_project, new_status='in-progress')
    print()
    _update_analyses(output_project, new_status='completed')
    print()

    print('Add/update analyses, gvcf -> joint_calling')
    _submit_analyses(samples, output_project, a_type='joint-calling')
    print()
    _update_analyses(output_project, new_status='in-progress')
    print()
    _update_analyses(output_project, new_status='completed')
    print()

    # Checking that after all calls, a 'completed' 'joint-calling' analysis must exist
    # that includes all initally added samples
    analysis = aapi.get_latest_complete_analysis_for_type(
        project=output_project, analysis_type=AnalysisType('joint-calling')
    )
    assert analysis['type'] == 'joint-calling'
    assert set([s['id'] for s in samples]).issubset(set(analysis['sample_ids']))


if __name__ == '__main__':
    test_simulate_joint_calling_pipeline(
        input_project=INPUT_PROJECT,
        output_project=OUTPUT_PROJECT,
    )



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.participant import (
    ParticipantLayer,
    ParticipantUpsert,
    ParticipantUpsertBody,
)
from db.python.layers.sample import SampleUpsert, SequenceUpsert
from models.enums import SequenceType, SampleType, SequenceStatus


class TestUpsert(DbIsolatedTest):
    """
    Test upsert functionality in SM
    """

    @run_as_sync
    async def test_insert_participants(self):
        """
        Test inserting participants, samples and sequences, and make sure they're correctly linked.

        Tests the other side of:
            tests.test_parse_generic_metadata:TestParseGenericMetadata.test_rows_with_participants
        """
        all_participants = [
            ParticipantUpsert.construct(
                **{
                    'external_id': 'Demeter',
                    'meta': {},
                    'samples': [
                        SampleUpsert.construct(
                            **{
                                'external_id': 'sample_id001',
                                'meta': {},
                                'sequences': [
                                    SequenceUpsert.construct(
                                        **{
                                            'meta': {
                                                'reads': [
                                                    [
                                                        {
                                                            'basename': 'sample_id001.filename-R1.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id001.filename-R1.fastq.gz',
                                                            'size': 111,
                                                        },
                                                        {
                                                            'basename': 'sample_id001.filename-R2.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id001.filename-R2.fastq.gz',
                                                            'size': 111,
                                                        },
                                                    ]
                                                ],
                                                'reads_type': 'fastq',
                                            },
                                            'status': SequenceStatus('uploaded'),
                                            'type': SequenceType('genome'),
                                        }
                                    ),
                                    SequenceUpsert.construct(
                                        **{
                                            'meta': {
                                                'reads': [
                                                    [
                                                        {
                                                            'basename': 'sample_id001.exome.filename-R1.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id001.exome.filename-R1.fastq.gz',
                                                            'size': 111,
                                                        },
                                                        {
                                                            'basename': 'sample_id001.exome.filename-R2.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id001.exome.filename-R2.fastq.gz',
                                                            'size': 111,
                                                        },
                                                    ]
                                                ],
                                                'reads_type': 'fastq',
                                            },
                                            'status': SequenceStatus('uploaded'),
                                            'type': SequenceType('exome'),
                                        }
                                    ),
                                ],
                                'type': SampleType('blood'),
                            }
                        )
                    ],
                }
            ),
            ParticipantUpsert.construct(
                **{
                    'external_id': 'Apollo',
                    'meta': {},
                    'samples': [
                        SampleUpsert.construct(
                            **{
                                'external_id': 'sample_id002',
                                'meta': {},
                                'sequences': [
                                    SequenceUpsert.construct(
                                        **{
                                            'meta': {
                                                'reads': [
                                                    [
                                                        {
                                                            'basename': 'sample_id002.filename-R1.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id002.filename-R1.fastq.gz',
                                                            'size': 111,
                                                        },
                                                        {
                                                            'basename': 'sample_id002.filename-R2.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id002.filename-R2.fastq.gz',
                                                            'size': 111,
                                                        },
                                                    ]
                                                ],
                                                'reads_type': 'fastq',
                                            },
                                            'status': SequenceStatus('uploaded'),
                                            'type': SequenceType('genome'),
                                        }
                                    )
                                ],
                                'type': SampleType('blood'),
                            }
                        ),
                        SampleUpsert.construct(
                            **{
                                'external_id': 'sample_id004',
                                'meta': {},
                                'sequences': [
                                    SequenceUpsert.construct(
                                        **{
                                            'meta': {
                                                'reads': [
                                                    [
                                                        {
                                                            'basename': 'sample_id004.filename-R1.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id004.filename-R1.fastq.gz',
                                                            'size': 111,
                                                        },
                                                        {
                                                            'basename': 'sample_id004.filename-R2.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id004.filename-R2.fastq.gz',
                                                            'size': 111,
                                                        },
                                                    ]
                                                ],
                                                'reads_type': 'fastq',
                                            },
                                            'status': SequenceStatus('uploaded'),
                                            'type': SequenceType('genome'),
                                        }
                                    )
                                ],
                                'type': SampleType('blood'),
                            }
                        ),
                    ],
                }
            ),
            ParticipantUpsert.construct(
                **{
                    'external_id': 'Athena',
                    'meta': {},
                    'samples': [
                        SampleUpsert.construct(
                            **{
                                'external_id': 'sample_id003',
                                'meta': {},
                                'sequences': [
                                    SequenceUpsert.construct(
                                        **{
                                            'meta': {
                                                'reads': [
                                                    [
                                                        {
                                                            'basename': 'sample_id003.filename-R1.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id003.filename-R1.fastq.gz',
                                                            'size': 111,
                                                        },
                                                        {
                                                            'basename': 'sample_id003.filename-R2.fastq.gz',
                                                            'checksum': None,
                                                            'class': 'File',
                                                            'location': '/path/to/sample_id003.filename-R2.fastq.gz',
                                                            'size': 111,
                                                        },
                                                    ]
                                                ],
                                                'reads_type': 'fastq',
                                            },
                                            'status': SequenceStatus('uploaded'),
                                            'type': SequenceType('genome'),
                                        }
                                    )
                                ],
                                'type': SampleType('blood'),
                            }
                        )
                    ],
                }
            ),
        ]

        body = ParticipantUpsertBody.construct(participants=all_participants)
        # Table interfaces
        pt = ParticipantLayer(self.connection)

        await pt.batch_upsert_participants(body)

        expected_sample_eid_to_participant_eid = {
            sample.external_id: participant.external_id
            for participant in all_participants
            for sample in participant.samples
        }

        db_participants = await self.connection.connection.fetch_all(
            'SELECT * FROM participant'
        )
        self.assertEqual(3, len(db_participants))
        self.assertEqual('Demeter', db_participants[0]['external_id'])
        self.assertEqual('Apollo', db_participants[1]['external_id'])
        self.assertEqual('Athena', db_participants[2]['external_id'])

        participant_id_map = {p['external_id']: p['id'] for p in db_participants}

        db_samples = await self.connection.connection.fetch_all('SELECT * FROM sample')
        self.assertEqual(4, len(db_samples))
        for db_sample in db_samples:
            self.assertIsNotNone(db_sample['external_id'])
            self.assertIsNotNone(db_sample['participant_id'])
            # get expected_participant_id from the db_sample external_id
            expected_participant_eid = expected_sample_eid_to_participant_eid.get(
                db_sample['external_id']
            )
            self.assertEqual(
                participant_id_map[expected_participant_eid],
                db_sample['participant_id'],
            )



"""
Adds samples for a subsequent test_joint_calling_workflow.py test

Requires either write access to test_input_project,
or the server run with SM_ALLOWALLACCESS=1
"""

import random
import string

from sample_metadata.apis import SampleApi
from sample_metadata.models import NewSample, SampleType

INPUT_PROJECT = 'test_input_project'


sapi = SampleApi()


def _add_samples(run_id: str, project: str):
    """
    Add 3 samples: one with fastq input, one with CRAM input, one with GVCF input.
    :param run_id: to suffix sample names for uniqueness
    """
    s1 = NewSample(
        external_id=f'NA12878-from-fq-{run_id}',
        type=SampleType('blood'),
        meta={
            'reads': [
                [
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L001_R1.fq',
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L002_R1.fq',
                ],
                [
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L001_R2.fq',
                    'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878_L002_R2.fq',
                ],
            ]
        },
    )
    s2 = NewSample(
        external_id=f'NA12878-from-cram-{run_id}',
        type=SampleType('blood'),
        meta={'reads': 'gs://cpg-seqr-test/batches/NA12878-trio-tiny/NA12878.cram'},
    )
    s3 = NewSample(
        external_id=f'NA12878-from-gvcf-{run_id}',
        type=SampleType('blood'),
        meta={'reads': 'gs://cpg-seqr-test/batches/NA12878-trio/NA12878.g.vcf.gz'},
    )
    sample_ids = [sapi.create_new_sample(project, s) for s in (s1, s2, s3)]
    print(f'Added samples {", ".join(sample_ids)}')
    return sample_ids


if __name__ == '__main__':
    # Unique test run ID to avoid clashing with previous test run samples
    TEST_RUN_ID = ''.join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(6)
    )
    print(f'Test run ID: {TEST_RUN_ID}')
    print(f'Populate samples for test run {TEST_RUN_ID}, input project {INPUT_PROJECT}')
    _add_samples(TEST_RUN_ID, INPUT_PROJECT)
